﻿using Buzz.MachineInterface;
using BuzzGUI.Common;
using BuzzGUI.Interfaces;
using DSPLib.SoundTouch;
using LibSampleRate;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

namespace WDE.AudioBlock
{
    [MachineDecl(Name = "WDE AudioBlock", ShortName = "AudioBlock", Author = "WDE", MaxTracks = 1)]
    public class AudioBlock : IBuzzMachine
    {
        // Some defines used by AudioBlock
        private string AUDIO_BLOCK_VERSION = "1.0.4.2";

        public const int NUMBER_OF_AUDIO_BLOCKS = 16;
        public const int CLICK_REMOVAL_SAMPLES = 200;
        public const string AUDIO_BLOCK_PATTERN_BASE_NAME = "Audio ";
        public const double WAVE_CANVAS_WIDTH = 200;
        public const double WAVE_CANVAS_HEIGTH = 1000;
        public const double AUDIO_BLOCK_GUI_PARAM_WIDTH = 640;
        public const double AUDIO_BLOCK_GUI_WAVE_VIEW_MIN_WIDTH = 900;
        private const int PATTERN_LENGHT_STEP = 8;
        public IBuzzMachineHost host;
        public static object syncLock = new object();

        public const int NUMBER_OF_SPLINE_POINTS_PER_SECOND = 100;
        public SplineCache[] splineCacheVol;
        public SplineCache[] splineCachePan;

        private AboutWindow aboutWindow;

        public WaveUndo WaveUndo { get; set; }

        public enum EEnvelopeType
        {
            Volume,
            Pan
        }

        public AudioBlock(IBuzzMachineHost host)
        {
            this.host = host;
            Global.Buzz.Song.SequenceAdded += Song_SequenceAdded;
            Global.Buzz.Song.MachineAdded += Song_MachineAdded;
            Global.Buzz.OpenSong += Buzz_OpenSong;
            MachineInitialized = false;

            WaveUndo = new WaveUndo(1, this);

            splineCacheVol = new SplineCache[NUMBER_OF_AUDIO_BLOCKS];
            splineCachePan = new SplineCache[NUMBER_OF_AUDIO_BLOCKS];
        }


        /// <summary>
        /// Do some initialization once song has been loaded.
        /// </summary>
        /// <param name="obj"></param>
        private void Buzz_OpenSong(IOpenSong obj)
        {
            ReCreateEnvPoints();
        }

        /// <summary>
        /// Add some event handling once the machine has been laoded.
        /// </summary>
        /// <param name="obj"></param>
        private void Song_MachineAdded(IMachine obj)
        {
            // host is available here.
            if (host.Machine == obj)
            {
                host.Machine.Graph.Buzz.Song.Wavetable.WaveChanged += Wavetable_WaveChanged;
                host.Machine.PatternAdded += Machine_PatternAdded;
                host.Machine.PatternRemoved += Machine_PatternRemoved;
            }
        }

        /// <summary>
        /// Sequece was added to the somg.
        /// </summary>
        /// <param name="obj"></param>
        private void Song_SequenceAdded(int obj)
        {
            // At this point the machines and patterns are loaded(?), but not sequences.

            // If only one pattern --> rename
            if (host.Machine.Patterns.Count == 1)
                if (host.Machine.Patterns[0].Name == "00")
                {
                    string name = AUDIO_BLOCK_PATTERN_BASE_NAME + "00";
                    host.Machine.DeletePattern(host.Machine.Patterns[0]);
                    host.Machine.CreatePattern(name, 16);
                    MachineState.AudioBlockInfoTable[0].Pattern = name;
                }
            MachineInitialized = true;
        }

        /// <summary>
        /// Update UI if pattern removed.
        /// </summary>
        /// <param name="obj"></param>
        private void Machine_PatternRemoved(IPattern obj)
        {
            UpdateUI();
        }

        /// <summary>
        /// Update UI if pattern added.
        /// </summary>
        /// <param name="obj"></param>
        private void Machine_PatternAdded(IPattern obj)
        {
            UpdateUI();
        }

        /// <summary>
        /// Called when somethingchanges in Wavetable. Note that this is also called when song is loading, so this event is rised both from user action or from loading.
        /// </summary>
        /// <param name="obj"></param>
        private void Wavetable_WaveChanged(int obj)
        {
            // Find out if there was a Audio Block linked to that wave and clear them.
            var wt = host.Machine.Graph.Buzz.Song.Wavetable;

            if (wt.Waves[obj] == null) // Removed
            {
                for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
                {
                    if (MachineState.AudioBlockInfoTable[i].WavetableIndex == obj)
                    {
                        MachineState.AudioBlockInfoTable[i].WavetableIndex = -1;
                        ResetVolEnvPoints(i);
                        ResetPanEnvPoints(i);
                    }
                }
            }
            else if (MachineInitialized) // Machine has been initialized so loading new wave
            {
                for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
                {
                    if (MachineState.AudioBlockInfoTable[i].WavetableIndex == obj)
                    {
                        UpdateLoopRange(i);
                        SetPatternLength(i);
                        UpdateEnvData(i);
                        RaiseUpdateWaveGraphEvent(i, WaveUpdateEventType.NewWaveAdded);
                        RefreshBuzzViews();
                    }
                }
            }
            UpdateUI();
        }

        public void UpdateLoopRange(int audioBlockIndex)
        {
            lock (syncLock)
            {
                AudioBlockInfo abi = MachineState.AudioBlockInfoTable[audioBlockIndex];
                var wt = host.Machine.Graph.Buzz.Song.Wavetable;
                double loopStartInSeconds = wt.Waves[abi.WavetableIndex].Layers[0].LoopStart / (double)wt.Waves[abi.WavetableIndex].Layers[0].SampleRate;
                double offsetInSeconds = abi.OffsetInMs * 1000 + abi.OffsetInSeconds;

                if (abi.LoopEnabled && loopStartInSeconds < offsetInSeconds)
                {
                    abi.OffsetInMs = 0;
                    abi.OffsetInSeconds = 0;
                }
            }
        }

        /// <summary>
        /// Updates spline data and makes some sanity checks os that envelope points do not go below/above max/min possible values.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        public void UpdateEnvData(int audioBlockIndex)
        {
            UpdatePanEnvPoints(audioBlockIndex);
            UpdateVolumeEnvPoints(audioBlockIndex);
            UpdateSplinePan(audioBlockIndex);
            UpdateSplineVol(audioBlockIndex);
        }

        /// <summary>
        /// This is one way to refres/invalidate Buzz views...
        /// </summary>
        public void RefreshBuzzViews()
        {
            if (host.Machine.Graph.Buzz.ActiveView == BuzzView.SequenceView)
            {
                host.Machine.Graph.Buzz.ActiveView = BuzzView.PatternView;
                host.Machine.Graph.Buzz.ActiveView = BuzzView.SequenceView; // update view;
            }
            else if (host.Machine.Graph.Buzz.ActiveView == BuzzView.PatternView)
            {
                host.Machine.Graph.Buzz.ActivatePatternEditor();
            }
        }

        /// <summary>
        /// Call this when data has changed and UI needs to be updated.
        /// </summary>
        private void UpdateUI()
        {
            if (AudioBlockGUI != null)
                AudioBlockGUI.UpdateUI();
        }

        [ParameterDecl(MaxValue = 1, DefValue = 0, ValueDescriptions = new String[] { "No", "Yes" })]
        public int Mute { get; set; }

        enum EnumMute : int
        {
            No,
            Yes
        }

        /// <summary>
        /// Work method is called by Buzz and here we generate the sound. Try to use as small buffers as possible. Use locks to make sure no other thread is touching wavetable data.
        /// </summary>
        /// <param name="output"></param>
        /// <param name="numsamples"></param>
        /// <param name="mode"></param>
        /// <returns></returns>
        public bool Work(Sample[] output, int numsamples, WorkModes mode)
        {
            bool ret = false;

            if (this.Mute == (int)EnumMute.Yes)
                return ret;

            if (Global.Buzz.Playing || Global.Buzz.Recording)
            {
                string patternName = "";

                for (int i = 0; i < numsamples; i++)
                {
                    output[i].L = 0.0f;
                    output[i].R = 0.0f;
                }
                ReadOnlyCollection<ISequence> seqs = GetPlayingSequences(host.Machine);
                lock (syncLock)
                {
                    foreach (ISequence seq in seqs)
                    {
                        int playPosition = GetPositionInPattern(out patternName, seq);

                        if (playPosition != -1)
                        {
                            float gain = GetGainForPattern(patternName);
                            gain *= (float)GetVolumeEnvelopeGain(patternName, playPosition);
                            Tuple<double, double> panGain = GetPanEnvelopeGain(patternName, playPosition);
                            float gainL = gain * (float)panGain.Item1;
                            float gainR = gain * (float)panGain.Item2;
                            ret |= FetchAudioFromWavetableWithSampleRateCorrection(playPosition, patternName, numsamples, ref output, gainL, gainR, true);
                        }
                    }

                    // If playing in pattern editor
                    if (seqs.Count == 0)
                    {
                        int playPosition = GetPositionInPlayingPatternEditorPattern(out patternName);
                        if (playPosition != -1)
                        {
                            float gain = GetGainForPattern(patternName);
                            gain *= (float)GetVolumeEnvelopeGain(patternName, playPosition);
                            Tuple<double, double> panGain = GetPanEnvelopeGain(patternName, playPosition);
                            float gainL = gain * (float)panGain.Item1;
                            float gainR = gain * (float)panGain.Item2;
                            ret |= FetchAudioFromWavetableWithSampleRateCorrection(playPosition, patternName, numsamples, ref output, gainL, gainR, true);
                        }
                    }
                }
            }
            else
                ret = false;

            return ret;
        }


        /// <summary>
        /// Enable/disable audio looping.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <param name="enabled"></param>
        internal void SetLoopingState(int audioBlockIndex, bool enabled)
        {
            lock (syncLock)
            {
                // Check bounds
                var wt = host.Machine.Graph.Buzz.Song.Wavetable;
                AudioBlockInfo abi = MachineState.AudioBlockInfoTable[audioBlockIndex];
                double loopStartInSeconds = wt.Waves[abi.WavetableIndex].Layers[0].LoopStart / (double)wt.Waves[abi.WavetableIndex].Layers[0].SampleRate;

                // Don't go to the wrong side of loop beginning
                if (enabled && (loopStartInSeconds < abi.OffsetInSeconds + abi.OffsetInMs / 1000.0))
                {
                    abi.OffsetInSeconds = 0;
                    abi.OffsetInMs = 0;
                    UpdateUI();
                }
                MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled = enabled;
            }
        }

        /// <summary>
        /// Open loop is an another way of looping samples. Not used currently.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <param name="v"></param>
        internal void SetOpenLoopState(int audioBlockIndex, bool v)
        {
            lock (syncLock)
            {
                MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop = v;
            }
        }

        /// <summary>
        /// Resturs volume gain from curves or lines. Returs 1.0f is envelopes are notused.
        /// </summary>
        /// <param name="patternName"></param>
        /// <param name="playPosition"></param>
        /// <returns></returns>
        private double GetVolumeEnvelopeGain(string patternName, int playPosition)
        {
            double ret = 1.0f;

            AudioBlockInfo abiFound = null;
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
            {
                if (abi.Pattern == patternName)
                {
                    abiFound = abi;
                    break;
                }
            }

            if (abiFound != null && abiFound.VolumeEnvelopePoints.Count >= 2 && abiFound.VolumeEnvelopeEnabled)
            {
                if (abiFound.VolumeEnvelopeCurves)
                {
                    ret = CalculateValueFromEnvelopeCurve(abiFound, playPosition, patternName, EEnvelopeType.Volume);
                }
                else
                {
                    List<IEnvelopePoint> volPointListLine = abiFound.VolumeEnvelopePoints.ConvertAll<IEnvelopePoint>(x => (IEnvelopePoint)x);
                    ret = CalculateValueFromEnvelopeLine(patternName, playPosition, volPointListLine);
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns value from specific point in spline.
        /// </summary>
        /// <param name="abi"></param>
        /// <param name="playPosition"></param>
        /// <param name="patternName"></param>
        /// <param name="type"></param>
        /// <returns></returns>
        private double CalculateValueFromEnvelopeCurve(AudioBlockInfo abi, int playPosition, string patternName, EEnvelopeType type)
        {
            double ret = 1.0;
            int index = -1;
            for (int i = 0; i < MachineState.AudioBlockInfoTable.Length; i++)
                if (MachineState.AudioBlockInfoTable[i] == abi)
                {
                    index = i;
                    break;
                }

            if (index >= 0)
            {
                SplineCache splineCache;
                if (type == EEnvelopeType.Volume)
                {
                    splineCache = GetSplineVol(index);
                }
                else
                {
                    splineCache = GetSplinePan(index);

                }

                float[] yValues = splineCache.YValues;
                double playPositionInSeconds = playPosition / (double)Global.Buzz.SelectedAudioDriverSampleRate;

                playPositionInSeconds = UpdatePlayPosInSecondsIfLooped(playPositionInSeconds, GetWavetableIndexForPattern(patternName), patternName);

                double stepSizeInSeconds = splineCache.StepSizeInSeconds;

                if (stepSizeInSeconds > 0)
                {

                    int startPos = (int)(playPositionInSeconds / stepSizeInSeconds);

                    if (startPos >= yValues.Length - 1)
                    {
                        ret = yValues[yValues.Length - 1];
                    }
                    else
                    {
                        double posInStep = playPositionInSeconds - stepSizeInSeconds * (double)startPos;
                        double mul = (yValues[startPos + 1] - yValues[startPos]) / stepSizeInSeconds;
                        ret = yValues[startPos] + mul * posInStep;
                    }
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns spline volume cache for AudioBlock
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <returns></returns>
        internal SplineCache GetSplineVol(int audioBlockIndex)
        {
            return splineCacheVol[audioBlockIndex];
        }

        /// <summary>
        /// Returns spline pan cache for AudioBlock
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <returns></returns>
        internal SplineCache GetSplinePan(int audioBlockIndex)
        {
            return splineCachePan[audioBlockIndex];
        }

        /// <summary>
        /// Resturs pan from curves or lines. Returs 1.0f for both left and right if envelopes are not used. User power panning algorithm.
        /// </summary>
        /// <param name="patternName"></param>
        /// <param name="playPosition"></param>
        /// <returns></returns>
        private Tuple<double, double> GetPanEnvelopeGain(string patternName, int playPosition)
        {
            Tuple<double, double> ret = new Tuple<double, double>(1.0, 1.0);

            AudioBlockInfo abiFound = null;
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
            {
                if (abi.Pattern == patternName)
                {
                    abiFound = abi;
                    break;
                }
            }

            if (abiFound != null && abiFound.PanEnvelopePoints.Count >= 2 && abiFound.PanEnvelopeEnabled)
            {
                double pan = 1.0;

                if (abiFound.PanEnvelopeCurves)
                {
                    pan = CalculateValueFromEnvelopeCurve(abiFound, playPosition, patternName, EEnvelopeType.Pan);
                }
                else
                {
                    List<IEnvelopePoint> abiFondI = abiFound.PanEnvelopePoints.ConvertAll<IEnvelopePoint>(x => (IEnvelopePoint)x);
                    pan = CalculateValueFromEnvelopeLine(patternName, playPosition, abiFondI); // 0-2
                }
                pan = (pan / 2.0) * Math.PI / 2.0;
                ret = new Tuple<double, double>(2.0 / Math.Sqrt(2.0) * Math.Cos(pan), 2.0 / Math.Sqrt(2) * Math.Sin(pan));

            }
            return ret;
        }

        /// <summary>
        /// Returns value from specific point in envelope line.
        /// </summary>
        /// <param name="patternName"></param>
        /// <param name="playPosition"></param>
        /// <param name="envelopePoints"></param>
        /// <returns></returns>
        internal double CalculateValueFromEnvelopeLine(string patternName, int playPosition, List<IEnvelopePoint> envelopePoints)
        {
            double ret;

            double playPositionInSeconds = playPosition / (double)Global.Buzz.SelectedAudioDriverSampleRate;

            playPositionInSeconds = UpdatePlayPosInSecondsIfLooped(playPositionInSeconds, GetWavetableIndexForPattern(patternName), patternName);

            IEnvelopePoint prevPoint = envelopePoints[0];

            ret = prevPoint.Value;

            bool found = false;
            for (int i = 1; i < envelopePoints.Count; i++)
            {
                if (prevPoint.TimeStamp <= playPositionInSeconds && playPositionInSeconds < envelopePoints[i].TimeStamp)
                {
                    found = true;
                    double lenght = envelopePoints[i].TimeStamp - prevPoint.TimeStamp;
                    double timeX = playPositionInSeconds - prevPoint.TimeStamp;
                    double mul = (envelopePoints[i].Value - prevPoint.Value) / lenght;
                    ret = prevPoint.Value + timeX * mul;
                    break;
                }
                prevPoint = envelopePoints[i];
            }
            if (!found)
                ret = envelopePoints[envelopePoints.Count - 1].Value;

            return ret;
        }


        /// <summary>
        /// Reads samples from wavetable based on position in AudioBlock pattern. Takes into account diffrerent Buzz sample rates, and wavetable samplerates.
        /// </summary>
        /// <param name="playPosition"></param>
        /// <param name="patternName"></param>
        /// <param name="numsamples"></param>
        /// <param name="output"></param>
        /// <param name="gainL"></param>
        /// <param name="gainR"></param>
        /// <param name="scaleUp"></param>
        /// <returns></returns>
        public bool FetchAudioFromWavetableWithSampleRateCorrection(int playPosition, string patternName, int numsamples, ref Sample[] output, float gainL, float gainR, bool scaleUp)
        {
            double playPositionInSample = (double)playPosition;

            bool ret = false;

            int sampleIndex = GetWavetableIndexForPattern(patternName);

            lock (syncLock)
            {
                if (sampleIndex >= 0)
                {
                    var wt = host.Machine.Graph.Buzz.Song.Wavetable;

                    double step = GetSampleFrequency(sampleIndex) / (double)Global.Buzz.SelectedAudioDriverSampleRate;

                    Sample[] outputTmp = new Sample[(int)(Math.Ceiling(numsamples * step))];
                    double offset = step * GetOffsetForPatternInMs(patternName) / 1000.0 * ((double)Global.Buzz.SelectedAudioDriverSampleRate);

                    playPositionInSample *= step;
                    playPositionInSample += offset;

                    // Update loop position
                    playPositionInSample = UpdatePlayPosIfLooped(playPositionInSample, sampleIndex, patternName);

                    #region ADD_SILENCE
                    // Add silence if offset
                    int outputOffset = 0;

                    if (playPositionInSample < 0)
                    {
                        for (int i = 0; i < numsamples; i++)
                        {
                            playPositionInSample += step;
                            outputOffset++;
                            if (playPositionInSample >= 0)
                                break;
                        }
                    }
                    #endregion

                    GetSamples(sampleIndex, ref outputTmp, outputTmp.Length, (int)playPositionInSample, scaleUp);

                    #region CLICK_REMOVAL
                    // Very basic click removal if using offset (cut from beginning)
                    if (playPosition == 0 && offset > 0)
                    {
                        int end = CLICK_REMOVAL_SAMPLES;
                        end = end > numsamples ? numsamples : end;
                        SmoothCut(ref outputTmp, 0, end);
                    }
                    #endregion

                    // Need to check if we are looping
                    double loopEnd = 0;
                    double loopStart = 0;
                    bool looping = IsLoopingPattern(patternName);
                    if (looping)
                    {
                        loopEnd = wt.Waves[sampleIndex].Layers[0].LoopEnd;
                        loopStart = wt.Waves[sampleIndex].Layers[0].LoopStart;
                    }
                    double stepper = 0;

                    int copyIndex;
                    int samplesToCopy = numsamples - outputOffset;
                    for (copyIndex = 0; copyIndex < samplesToCopy; copyIndex++)
                    {
                        if (looping && (playPositionInSample + stepper >= loopEnd))
                        {
                            // There are still samples to copy but we have reached the end of the loop.
                            break;
                        }

                        output[copyIndex + outputOffset].L += outputTmp[(int)stepper].L * gainL;
                        output[copyIndex + outputOffset].R += outputTmp[(int)stepper].R * gainR;
                        stepper += step;
                    }

                    // If still samples to copy in the beginning of the loop                
                    if (copyIndex < samplesToCopy)
                    {
                        stepper = 0;
                        playPositionInSample = loopStart;
                        GetSamples(sampleIndex, ref outputTmp, outputTmp.Length, (int)playPositionInSample, scaleUp);

                        for (; copyIndex < samplesToCopy; copyIndex++)
                        {
                            output[copyIndex + outputOffset].L += outputTmp[(int)stepper].L * gainL;
                            output[copyIndex + outputOffset].R += outputTmp[(int)stepper].R * gainR;
                            stepper += step;
                        }
                    }

                    ret = true;
                }
            }
            return ret;
        }


        /// <summary>
        /// Returns updated play position if sample is looping.
        /// </summary>
        /// <param name="playPositionInSample"></param>
        /// <param name="sampleIndex"></param>
        /// <param name="patternName"></param>
        /// <returns></returns>
        internal double UpdatePlayPosIfLooped(double playPositionInSample, int sampleIndex, string patternName)
        {
            double ret = playPositionInSample;

            if (IsLoopingPattern(patternName))
            {
                lock (syncLock)
                {
                    var wt = host.Machine.Graph.Buzz.Song.Wavetable;

                    double loopStart = wt.Waves[sampleIndex].Layers[0].LoopStart;
                    double loopEnd = wt.Waves[sampleIndex].Layers[0].LoopEnd;
                    double loopLenght = loopEnd - loopStart;

                    // Are we in the loop?
                    if (playPositionInSample > loopStart)
                    {
                        ret = ((playPositionInSample - (loopStart)) % loopLenght) + loopStart;
                    }
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns updated play position in seconds if looping enabled.
        /// </summary>
        /// <param name="playPositionInSeconds"></param>
        /// <param name="sampleIndex"></param>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public double UpdatePlayPosInSecondsIfLooped(double playPositionInSeconds, int sampleIndex, string patternName)
        {
            double ret = playPositionInSeconds;

            if (IsLoopingPattern(patternName) && !IsOpenLoop(patternName))
            {
                lock (syncLock)
                {
                    var wt = host.Machine.Graph.Buzz.Song.Wavetable;

                    double loopStart = (wt.Waves[sampleIndex].Layers[0].LoopStart / (double)wt.Waves[sampleIndex].Layers[0].SampleRate) - GetOffsetForPatternInMs(patternName) / 1000.0;
                    double loopEnd = (wt.Waves[sampleIndex].Layers[0].LoopEnd / (double)wt.Waves[sampleIndex].Layers[0].SampleRate) - GetOffsetForPatternInMs(patternName) / 1000.0;
                    double loopLenght = loopEnd - loopStart;

                    // Are we in the loop?
                    if (playPositionInSeconds > loopStart)
                    {
                        ret = ((playPositionInSeconds - (loopStart)) % loopLenght) + loopStart;
                    }
                }
            }
            return ret;
        }

        /// <summary>
        /// Update offset for AudioBlock item. Can be sielnce (negative value) or cut (positive value).
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <param name="offsetInSeconds"></param>
        /// <param name="offsetInMs"></param>
        /// <returns></returns>
        internal bool UpdateOffsets(int audioBlockIndex, double offsetInSeconds, double offsetInMs)
        {
            bool ret = false;
            AudioBlockInfo abi = MachineState.AudioBlockInfoTable[audioBlockIndex];
            double changeDeltaInSeconds = offsetInSeconds - abi.OffsetInSeconds
                + (offsetInMs - abi.OffsetInMs) / 1000.0;

            if (abi.WavetableIndex == -1)
            {
                abi.OffsetInMs = 0;
                abi.OffsetInSeconds = 0;
                return false;
            }

            lock (syncLock)
            {
                var wt = host.Machine.Graph.Buzz.Song.Wavetable;

                double loopStartInSeconds = wt.Waves[abi.WavetableIndex].Layers[0].LoopStart / (double)wt.Waves[abi.WavetableIndex].Layers[0].SampleRate;

                // Don't go to the wrong side of loop beginning
                if (abi.LoopEnabled && (loopStartInSeconds < offsetInSeconds + offsetInMs / 1000.0))
                {
                    ret = false;
                }
                else
                {
                    ret = true;

                    abi.OffsetInSeconds = offsetInSeconds;
                    abi.OffsetInMs = offsetInMs;

                    List<VolumeEnvelopePoint> volPointList = abi.VolumeEnvelopePoints;

                    for (int i = 1; i < volPointList.Count; i++)
                    {
                        VolumeEnvelopePoint vep = volPointList[i];
                        vep.TimeStamp -= changeDeltaInSeconds;
                    }

                    for (int i = 1; i < volPointList.Count - 1; i++)
                    {
                        VolumeEnvelopePoint vep = volPointList[i];
                        if (vep.TimeStamp < 0)
                        {
                            volPointList.RemoveAt(i);
                            i--;
                        }
                    }


                    volPointList[volPointList.Count - 1].TimeStamp = GetSampleLengthInSecondsWithOffset(audioBlockIndex);

                    List<PanEnvelopePoint> panPointList = abi.PanEnvelopePoints;

                    for (int i = 1; i < panPointList.Count; i++)
                    {
                        PanEnvelopePoint vep = panPointList[i];
                        vep.TimeStamp -= changeDeltaInSeconds;
                    }

                    for (int i = 1; i < panPointList.Count - 1; i++)
                    {
                        PanEnvelopePoint vep = panPointList[i];
                        if (vep.TimeStamp < 0)
                        {
                            panPointList.RemoveAt(i);
                            i--;
                        }
                    }

                    panPointList[panPointList.Count - 1].TimeStamp = GetSampleLengthInSecondsWithOffset(audioBlockIndex);

                    UpdateEnvPoints(audioBlockIndex);
                    UpdateEnvData(audioBlockIndex);
                }
            }
            return ret;
        }

        /// <summary>
        /// Refresh/update envelope points and spline cache.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void UpdateEnvPoints(int audioBlockIndex)
        {
            UpdateVolumeEnvPoints(audioBlockIndex);
            UpdatePanEnvPoints(audioBlockIndex);
            if (MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeCurves)
                UpdateSplineVol(audioBlockIndex);
            if (MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopeCurves)
                UpdateSplinePan(audioBlockIndex);
        }

        /// <summary>
        /// Refres envelope data for all items.
        /// </summary>
        internal void ReCreateEnvPoints()
        {
            for (int i = 0; i < NUMBER_OF_AUDIO_BLOCKS; i++)
            {
                // Make sure everything is ok.
                UpdateEnvPoints(i);
            }
        }

        /// <summary>
        /// Sanity checks for Volume envelope.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void UpdateVolumeEnvPoints(int audioBlockIndex)
        {
            double maxLength = GetSampleLengthInSecondsWithOffset(audioBlockIndex);
            List<VolumeEnvelopePoint> volPoints = MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopePoints;
            if (volPoints.Count < 2)
            {
                volPoints.Clear();

                VolumeEnvelopePoint vep = new VolumeEnvelopePoint();
                vep.TimeStamp = 0;
                vep.Value = 1;
                volPoints.Add(vep);

                vep = new VolumeEnvelopePoint();
                vep.TimeStamp = maxLength;
                vep.Value = 1;
                volPoints.Add(vep);
            }

            // Update first and last
            volPoints[0].TimeStamp = 0;
            volPoints[volPoints.Count - 1].TimeStamp = maxLength;

            for (int i = 1; i < volPoints.Count - 1; i++)
            {
                VolumeEnvelopePoint vep = volPoints[i];
                if (vep.TimeStamp < 0)
                {
                    volPoints.RemoveAt(i);
                    i--;
                }
                if (vep.TimeStamp > maxLength)
                {
                    volPoints.RemoveAt(i);
                    i--;
                }
            }
        }

        /// <summary>
        /// Sanity checks for pan envelope.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void UpdatePanEnvPoints(int audioBlockIndex)
        {
            double maxLength = GetSampleLengthInSecondsWithOffset(audioBlockIndex);
            List<PanEnvelopePoint> panPoints = MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopePoints;
            if (panPoints.Count < 2)
            {
                panPoints.Clear();

                PanEnvelopePoint vep = new PanEnvelopePoint();
                vep.TimeStamp = 0;
                vep.Value = 1;
                panPoints.Add(vep);

                vep = new PanEnvelopePoint();
                vep.TimeStamp = maxLength;
                vep.Value = 1;
                panPoints.Add(vep);
            }

            // Update first and last
            panPoints[0].TimeStamp = 0;
            panPoints[panPoints.Count - 1].TimeStamp = maxLength;

            for (int i = 1; i < panPoints.Count - 1; i++)
            {
                PanEnvelopePoint vep = panPoints[i];
                if (vep.TimeStamp < 0)
                {
                    panPoints.RemoveAt(i);
                    i--;
                }
                if (vep.TimeStamp > maxLength)
                {
                    panPoints.RemoveAt(i);
                    i--;
                }
            }
        }

        /// <summary>
        /// Reset volume envelopes and clear previous data.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void ResetVolEnvPoints(int audioBlockIndex)
        {
            bool state = MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeCurves;

            MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeCurves = false;
            MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopePoints.Clear();
            UpdateSplineVol(audioBlockIndex);
            MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeCurves = state;
        }

        /// <summary>
        /// Reset pan envelopes and clear previous data.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void ResetPanEnvPoints(int audioBlockIndex)
        {
            bool state = MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopeCurves;

            MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopeCurves = false;
            MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopePoints.Clear();
            UpdateSplinePan(audioBlockIndex);
            MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopeCurves = state;
        }

        /// <summary>
        /// Is the pattern looping?
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        private bool IsLoopingPattern(string patternName)
        {
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
                if (abi.Pattern == patternName && abi.LoopEnabled)
                    return true;
            return false;
        }

        /// <summary>
        /// Is open loop mode?
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        private bool IsOpenLoop(string patternName)
        {
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
                if (abi.Pattern == patternName && abi.OpenLoop)
                    return true;
            return false;
        }

        /// <summary>
        /// Returns offset for pattern in milliseconds.
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public double GetOffsetForPatternInMs(string patternName)
        {
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
            {
                if (abi.Pattern == patternName)
                {
                    return abi.OffsetInMs + abi.OffsetInSeconds * 1000.0; // Milliseconds
                }
            }
            return 0;
        }

        /// <summary>
        /// Returns wavetable index of the associated wave for AudioBlock item.
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public int GetWavetableIndexForPattern(string patternName)
        {
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
            {
                if (abi.Pattern == patternName)
                {
                    return GetTableIndexFromWavetableIndex(abi.WavetableIndex);
                }
            }
            return -1;
        }

        /// <summary>
        /// Returns gain for AudioBlock item.
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public float GetGainForPattern(string patternName)
        {
            foreach (AudioBlockInfo abi in MachineState.AudioBlockInfoTable)
            {
                if (abi.Pattern == patternName)
                {
                    return abi.Gain;
                }
            }
            return 1.0f;
        }

        public int GetTableIndexFromWavetableIndex(int wavetableIndex)
        {
            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
            for (int i = 0; i < wt.Waves.Count; i++)
            {
                if (wt.Waves[i] != null && wt.Waves[i].Index == wavetableIndex)
                {
                    return i;
                }
            }
            return -1;
        }

        /// <summary>
        /// Update/recreate volume spline cache.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void UpdateSplineVol(int audioBlockIndex)
        {
            lock (syncLock)
            {
                if (MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeCurves)
                {
                    SplineCache spline = new SplineCache();
                    splineCacheVol[audioBlockIndex] = spline;

                    float min = 0.0f;
                    float max = 1.0f;
                    if (MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopeMode == (int)VolumeEnvelopeModes.dB6)
                        max = (float)Decibel.FromAmplitude(6.0);

                    List<IEnvelopePoint> volPoints = MachineState.AudioBlockInfoTable[audioBlockIndex].VolumeEnvelopePoints.ConvertAll<IEnvelopePoint>(x => (IEnvelopePoint)x);
                    spline.CreateSpline(volPoints.ToArray(), NUMBER_OF_SPLINE_POINTS_PER_SECOND);
                    spline.CutMinMax(min, max);
                }
            }
        }

        /// <summary>
        /// Update/recureate spline pan cache.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        internal void UpdateSplinePan(int audioBlockIndex)
        {
            lock (syncLock)
            {
                if (MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopeCurves)
                {
                    SplineCache spline = new SplineCache();
                    splineCachePan[audioBlockIndex] = spline;

                    float min = 0.0f;
                    float max = 2.0f;

                    List<IEnvelopePoint> envPoints = MachineState.AudioBlockInfoTable[audioBlockIndex].PanEnvelopePoints.ConvertAll<IEnvelopePoint>(x => (IEnvelopePoint)x);
                    spline.CreateSpline(envPoints.ToArray(), NUMBER_OF_SPLINE_POINTS_PER_SECOND);
                    spline.CutMinMax(min, max);
                }
            }
        }


        /// <summary>
        /// Get the actual sample data from wavetable.
        /// </summary>
        /// <param name="wabetableWaveIndex"></param>
        /// <param name="output"></param>
        /// <param name="numsamples"></param>
        /// <param name="positionInSample"></param>
        /// <param name="scaleUp"></param>
        public void GetSamples(int wabetableWaveIndex, ref Sample[] output, int numsamples, int positionInSample, bool scaleUp)
        {
            float[] tmpSamples = new float[numsamples * 10];

            lock (syncLock)
            {
                IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;

                if (wt.Waves.Count > wabetableWaveIndex && wt.Waves[wabetableWaveIndex].Layers[0] != null)
                {
                    if (wt.Waves[wabetableWaveIndex].Layers[0].SampleCount < positionInSample)
                        return;

                    if (wt.Waves[wabetableWaveIndex].Layers[0].SampleCount < positionInSample + numsamples)
                        numsamples = wt.Waves[wabetableWaveIndex].Layers[0].SampleCount - positionInSample;

                    if (numsamples > 0)
                    {
                        if (!IsStereo(wabetableWaveIndex))
                        {
                            wt.Waves[wabetableWaveIndex].Layers[0].GetDataAsFloat(tmpSamples, 0, 2, 0, positionInSample, numsamples);
                            wt.Waves[wabetableWaveIndex].Layers[0].GetDataAsFloat(tmpSamples, 1, 2, 0, positionInSample, numsamples);
                        }
                        else
                        {
                            wt.Waves[wabetableWaveIndex].Layers[0].GetDataAsFloat(tmpSamples, 0, 2, 0, positionInSample, numsamples);
                            wt.Waves[wabetableWaveIndex].Layers[0].GetDataAsFloat(tmpSamples, 1, 2, 1, positionInSample, numsamples);
                        }
                    }
                }
            }

            // We could also use volume settings for samples from Wavetable. Ignoring now.
            // scaleUp flag is used to avoid unncessary math operation if wave is for example drawn canvas.
            if (scaleUp)
            {
                for (int i = 0; i < numsamples; i++)
                {
                    output[i].L = tmpSamples[i * 2] * 32768.0f;
                    output[i].R = tmpSamples[(i * 2) + 1] * 32768.0f;
                }
            }
            else
            {
                for (int i = 0; i < numsamples; i++)
                {
                    output[i].L = tmpSamples[i * 2];
                    output[i].R = tmpSamples[(i * 2) + 1];
                }
            }
        }

        /// <summary>
        /// Resample audio in wavetable to Buzz samplerate using libsamplerate. Improves sound quality since AudioBlock does not need to do linear interpolatio from samples.
        /// </summary>
        /// <param name="wavetableIndex"></param>
        public void Resample(int wavetableIndex)
        {
            bool playing = Global.Buzz.Playing;
            Global.Buzz.Playing = false;

            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;

            double inputRate = wt.Waves[wavetableIndex].Layers[0].SampleRate;
            double outputRate = Global.Buzz.SelectedAudioDriverSampleRate;

            if (inputRate != outputRate)
            {
                int sampleCount = wt.Waves[wavetableIndex].Layers[0].SampleCount;
                string name = wt.Waves[wavetableIndex].Name;
                int rootNote = wt.Waves[wavetableIndex].Layers[0].RootNote;
                int loopStart = (int)Math.Round(wt.Waves[wavetableIndex].Layers[0].LoopStart * outputRate / inputRate);
                int loopEnd = (int)Math.Round(wt.Waves[wavetableIndex].Layers[0].LoopEnd * outputRate / inputRate);

                if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 1)
                {
                    float[] sourceData = new float[sampleCount];
                    float[] destData = new float[(int)(1000 + sampleCount * outputRate / inputRate)];
                    wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceData, 0, 1, 0, 0, sampleCount); // Mono

                    Resample(ref sourceData, ref destData, outputRate, inputRate);

                    wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, false, rootNote, false, true);

                    wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                    wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                    wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)outputRate;

                    wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Mono                    
                    wt.Waves[wavetableIndex].Layers[0].InvalidateData();
                }
                else if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 2)
                {
                    float[] sourceDataLeft = new float[sampleCount];
                    float[] sourceDataRight = new float[sampleCount];
                    float[] destData = new float[(int)(1000 + sampleCount * outputRate / inputRate)];
                    wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataLeft, 0, 1, 0, 0, sampleCount); // Left
                    wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataRight, 0, 1, 1, 0, sampleCount); // Right

                    wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, true, rootNote, false, true);

                    Resample(ref sourceDataLeft, ref destData, outputRate, inputRate);
                    sourceDataLeft = null;
                    wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Left 

                    Resample(ref sourceDataRight, ref destData, outputRate, inputRate);
                    wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 1, 0, destData.Length); // Right 

                    wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                    wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                    wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)outputRate;

                    wt.Waves[wavetableIndex].Layers[0].InvalidateData();
                }
            }

            Global.Buzz.Playing = playing;
        }

        /// <summary>
        /// Resample audio samples using libsamplerate.
        /// </summary>
        /// <param name="sourceData"></param>
        /// <param name="destData"></param>
        /// <param name="outputRate"></param>
        /// <param name="inputRate"></param>
        public void Resample(ref float[] sourceData, ref float[] destData, double outputRate, double inputRate)
        {

            var src = new SampleRateConverter(ConverterType.SRC_SINC_BEST_QUALITY, 1);
            src.SetRatio(outputRate / inputRate);

            int inputSampleCount;
            int outputSampleCount;
            int totalInputSampleCount = 0;
            int totalOutputSampleCount = 0;
            float[] inputBuffer = new float[1000];
            int inputBufferFillLevel = 0;
            int inputBufferReadOffset = 0;
            float[] outputBuffer = new float[1000];

            do
            {
                bool endOfInput = totalInputSampleCount == sourceData.Length;

                if (inputBufferFillLevel == 0)
                {
                    // Refill input buffer
                    inputBufferFillLevel = Math.Min(1000, sourceData.Length - totalInputSampleCount);
                    inputBufferReadOffset = 0;
                    Array.Copy(sourceData, totalInputSampleCount, inputBuffer, 0, inputBufferFillLevel);
                }

                src.Process(inputBuffer, inputBufferReadOffset, inputBufferFillLevel,
                    outputBuffer, 0, outputBuffer.Length,
                    endOfInput, out inputSampleCount, out outputSampleCount);

                inputBufferReadOffset += inputSampleCount;
                inputBufferFillLevel -= inputSampleCount;

                totalInputSampleCount += inputSampleCount;
                totalOutputSampleCount += outputSampleCount;

                Array.Copy(outputBuffer, 0, destData, totalOutputSampleCount, outputSampleCount);
            }
            while (inputSampleCount > 0 || outputSampleCount > 0);

        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="wavetableIndex"></param>
        /// <param name="semiTones"></param>
        public void ChangePitchSemitones(int wavetableIndex, float semiTones, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {
            bool playing = Global.Buzz.Playing;
            Global.Buzz.Playing = false;

            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;

            int sampleRate = wt.Waves[wavetableIndex].Layers[0].SampleRate;
            int channelCount = wt.Waves[wavetableIndex].Layers[0].ChannelCount;

            int sampleCount = wt.Waves[wavetableIndex].Layers[0].SampleCount;
            string name = wt.Waves[wavetableIndex].Name;
            int rootNote = wt.Waves[wavetableIndex].Layers[0].RootNote;
            int loopStart = wt.Waves[wavetableIndex].Layers[0].LoopStart;
            int loopEnd = wt.Waves[wavetableIndex].Layers[0].LoopEnd;

            if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 1)
            {
                float[] sourceData = new float[sampleCount];
                float[] destData = new float[(int)(1000 + sampleCount)];
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceData, 0, 1, 0, 0, sampleCount); // Mono

                ChangePitchSemitones(sampleRate, semiTones, ref sourceData, ref destData);

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, false, rootNote, false, true);

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Mono                    
                wt.Waves[wavetableIndex].Layers[0].InvalidateData();

            }
            else if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 2)
            {
                float[] sourceDataLeft = new float[sampleCount];
                float[] sourceDataRight = new float[sampleCount];
                float[] destData = new float[(int)(1000 + sampleCount)];

                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataLeft, 0, 1, 0, 0, sampleCount); // Left
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataRight, 0, 1, 1, 0, sampleCount); // Right

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, true, rootNote, false, true);

                ChangePitchSemitones(sampleRate, semiTones, ref sourceDataLeft, ref destData);

                sourceDataLeft = null;
                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Left 

                ChangePitchSemitones(sampleRate, semiTones, ref sourceDataRight, ref destData);

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 1, 0, destData.Length); // Right 

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].InvalidateData();
            }

            Global.Buzz.Playing = playing;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="soundTouch"></param>
        /// <param name="sourceDataLeft"></param>
        /// <param name="destData"></param>
        private void ChangePitchSemitones(int sampleRate, float semiTones, ref float[] sourceData, ref float[] destData, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {

            SoundTouch soundTouch = new SoundTouch();

            Global.Buzz.DCWriteLine(String.Format("SoundTouch Version {0}", soundTouch.VersionString));
            Global.Buzz.DCWriteLine(String.Format("Use QuickSeek: {0}", soundTouch.GetUseQuickSeek()));
            Global.Buzz.DCWriteLine(String.Format("Use AntiAliasing: {0}", soundTouch.GetUseAntiAliasing()));
            Global.Buzz.DCWriteLine(String.Format("AaFilterLength (ms): {0}", soundTouch.AaFilterLength()));
            Global.Buzz.DCWriteLine(String.Format("Overlap (ms): {0}", soundTouch.GetOverlapMs()));
            Global.Buzz.DCWriteLine(String.Format("Seek Window (ms): {0}", soundTouch.GetSeekWindowMs()));
            Global.Buzz.DCWriteLine(String.Format("Sequence (ms): {0}", soundTouch.GetSequenceMs()));

            soundTouch.SetSampleRate(sampleRate);
            soundTouch.SetChannels(1);
            soundTouch.SetPitchSemitones(semiTones);

            soundTouch.SetSequenceMs(sequenceMs);
            soundTouch.SetSeekWindowMs(seekWindowMs);
            soundTouch.SetOverlapMs(overlapMs);

            int totalInputSampleCount = 0;
            int totalOutputSampleCount = 0;
            float[] inputBuffer = new float[1000];
            int inputBufferFillLevel = 0;
            float[] outputBuffer = new float[1000];
            int outputBufferFillAmount = 0;

            do
            {
                bool endOfInput = totalInputSampleCount == sourceData.Length;

                if (inputBufferFillLevel == 0)
                {
                    // Refill input buffer
                    inputBufferFillLevel = Math.Min(1000, sourceData.Length - totalInputSampleCount);
                    Array.Copy(sourceData, totalInputSampleCount, inputBuffer, 0, inputBufferFillLevel);
                }

                soundTouch.PutSamples(inputBuffer, inputBufferFillLevel);
                totalInputSampleCount += inputBufferFillLevel;

                outputBufferFillAmount = Math.Min(1000, destData.Length - totalOutputSampleCount);
                outputBufferFillAmount = Math.Min(outputBufferFillAmount, soundTouch.NumberOfSamplesAvailable);
                outputBufferFillAmount = soundTouch.ReceiveSamples(outputBuffer, outputBufferFillAmount);

                Array.Copy(outputBuffer, 0, destData, totalOutputSampleCount, outputBufferFillAmount);

                totalOutputSampleCount += outputBufferFillAmount;

                inputBufferFillLevel = 0;
            }
            while (sourceData.Length - totalInputSampleCount > 0 || outputBufferFillAmount > 0);

            soundTouch.Dispose();

        }

        public void ChangeTempo(int wavetableIndex, float changePercent, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {
            bool playing = Global.Buzz.Playing;
            Global.Buzz.Playing = false;

            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
            int sampleRate = wt.Waves[wavetableIndex].Layers[0].SampleRate;
            int channelCount = wt.Waves[wavetableIndex].Layers[0].ChannelCount;

            int sampleCount = wt.Waves[wavetableIndex].Layers[0].SampleCount;
            string name = wt.Waves[wavetableIndex].Name;
            int rootNote = wt.Waves[wavetableIndex].Layers[0].RootNote;
            int loopStart = (int)(wt.Waves[wavetableIndex].Layers[0].LoopStart / (1.0 + changePercent / 100.0));
            int loopEnd = (int)(wt.Waves[wavetableIndex].Layers[0].LoopEnd / (1.0 + changePercent / 100.0));

            double targetBufferSize = sampleCount / (1.0 + changePercent / 100.0);

            if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 1)
            {
                float[] sourceData = new float[sampleCount];
                float[] destData = new float[(int)(1000 + targetBufferSize)];
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceData, 0, 1, 0, 0, sampleCount); // Mono

                ChangeTempo(sampleRate, changePercent, ref sourceData, ref destData);

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, false, rootNote, false, true);

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Mono                    
                wt.Waves[wavetableIndex].Layers[0].InvalidateData();
            }
            else if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 2)
            {
                float[] sourceDataLeft = new float[sampleCount];
                float[] sourceDataRight = new float[sampleCount];
                float[] destData = new float[(int)(1000 + targetBufferSize)];
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataLeft, 0, 1, 0, 0, sampleCount); // Left
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataRight, 0, 1, 1, 0, sampleCount); // Right

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, true, rootNote, false, true);

                ChangeTempo(sampleRate, changePercent, ref sourceDataLeft, ref destData);

                sourceDataLeft = null;
                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Left 

                ChangeTempo(sampleRate, changePercent, ref sourceDataRight, ref destData);

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 1, 0, destData.Length); // Right 

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].InvalidateData();
            }
            Global.Buzz.Playing = playing;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="soundTouch"></param>
        /// <param name="sourceDataLeft"></param>
        /// <param name="destData"></param>
        private void ChangeTempo(int sampleRate, float changePercent, ref float[] sourceData, ref float[] destData, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {

            SoundTouch soundTouch = new SoundTouch();

            Global.Buzz.DCWriteLine(String.Format("SoundTouch Version {0}", soundTouch.VersionString));
            Global.Buzz.DCWriteLine(String.Format("Use QuickSeek: {0}", soundTouch.GetUseQuickSeek()));
            Global.Buzz.DCWriteLine(String.Format("Use AntiAliasing: {0}", soundTouch.GetUseAntiAliasing()));

            soundTouch.SetSampleRate(sampleRate);
            soundTouch.SetChannels(1);
            soundTouch.SetTempoChange(changePercent);

            soundTouch.SetSequenceMs(sequenceMs);
            soundTouch.SetSeekWindowMs(seekWindowMs);
            soundTouch.SetOverlapMs(overlapMs);

            int totalInputSampleCount = 0;
            int totalOutputSampleCount = 0;
            float[] inputBuffer = new float[1000];
            int inputBufferFillLevel = 0;
            float[] outputBuffer = new float[1000];
            int outputBufferFillAmount = 0;

            do
            {
                bool endOfInput = totalInputSampleCount == sourceData.Length;

                if (inputBufferFillLevel == 0)
                {
                    // Refill input buffer
                    inputBufferFillLevel = Math.Min(1000, sourceData.Length - totalInputSampleCount);
                    Array.Copy(sourceData, totalInputSampleCount, inputBuffer, 0, inputBufferFillLevel);
                }

                soundTouch.PutSamples(inputBuffer, inputBufferFillLevel);
                totalInputSampleCount += inputBufferFillLevel;

                outputBufferFillAmount = Math.Min(1000, destData.Length - totalOutputSampleCount);
                outputBufferFillAmount = Math.Min(outputBufferFillAmount, soundTouch.NumberOfSamplesAvailable);
                outputBufferFillAmount = soundTouch.ReceiveSamples(outputBuffer, outputBufferFillAmount);

                Array.Copy(outputBuffer, 0, destData, totalOutputSampleCount, outputBufferFillAmount);

                totalOutputSampleCount += outputBufferFillAmount;

                inputBufferFillLevel = 0;
            }
            while (sourceData.Length - totalInputSampleCount > 0 || outputBufferFillAmount > 0);

            soundTouch.Dispose();

        }

        public void ChangeRate(int wavetableIndex, float changePercent, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {
            bool playing = Global.Buzz.Playing;
            Global.Buzz.Playing = false;

            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
            int sampleRate = wt.Waves[wavetableIndex].Layers[0].SampleRate;
            int channelCount = wt.Waves[wavetableIndex].Layers[0].ChannelCount;

            int sampleCount = wt.Waves[wavetableIndex].Layers[0].SampleCount;
            string name = wt.Waves[wavetableIndex].Name;
            int rootNote = wt.Waves[wavetableIndex].Layers[0].RootNote;
            int loopStart = (int)(wt.Waves[wavetableIndex].Layers[0].LoopStart / (1.0 + changePercent / 100.0));
            int loopEnd = (int)(wt.Waves[wavetableIndex].Layers[0].LoopEnd / (1.0 + changePercent / 100.0));

            double targetBufferSize = sampleCount / (1.0 + changePercent / 100.0);

            if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 1)
            {
                float[] sourceData = new float[sampleCount];
                float[] destData = new float[(int)(1000 + targetBufferSize)];
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceData, 0, 1, 0, 0, sampleCount); // Mono

                ChangeRate(sampleRate, changePercent, ref sourceData, ref destData, sequenceMs, seekWindowMs, overlapMs);

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, false, rootNote, false, true);

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Mono                    
                wt.Waves[wavetableIndex].Layers[0].InvalidateData();
            }
            else if (wt.Waves[wavetableIndex].Layers[0].ChannelCount == 2)
            {
                float[] sourceDataLeft = new float[sampleCount];
                float[] sourceDataRight = new float[sampleCount];
                float[] destData = new float[(int)(1000 + targetBufferSize)];
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataLeft, 0, 1, 0, 0, sampleCount); // Left
                wt.Waves[wavetableIndex].Layers[0].GetDataAsFloat(sourceDataRight, 0, 1, 1, 0, sampleCount); // Right

                wt.AllocateWave(wavetableIndex, "", name, destData.Length, WaveFormat.Float32, true, rootNote, false, true);

                ChangeRate(sampleRate, changePercent, ref sourceDataLeft, ref destData, sequenceMs, seekWindowMs, overlapMs);

                sourceDataLeft = null;
                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 0, 0, destData.Length); // Left 

                ChangeRate(sampleRate, changePercent, ref sourceDataRight, ref destData, sequenceMs, seekWindowMs, overlapMs);

                wt.Waves[wavetableIndex].Layers[0].SetDataAsFloat(destData, 0, 1, 1, 0, destData.Length); // Right 

                wt.Waves[wavetableIndex].Layers[0].LoopStart = loopStart;
                wt.Waves[wavetableIndex].Layers[0].LoopEnd = loopEnd;
                wt.Waves[wavetableIndex].Layers[0].SampleRate = (int)sampleRate;

                wt.Waves[wavetableIndex].Layers[0].InvalidateData();
            }

            Global.Buzz.Playing = false;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="soundTouch"></param>
        /// <param name="sourceDataLeft"></param>
        /// <param name="destData"></param>
        private void ChangeRate(int sampleRate, float changePercent, ref float[] sourceData, ref float[] destData, int sequenceMs = 0, int seekWindowMs = 0, int overlapMs = 8)
        {
            SoundTouch soundTouch = new SoundTouch();

            Global.Buzz.DCWriteLine(String.Format("SoundTouch Version {0}", soundTouch.VersionString));
            Global.Buzz.DCWriteLine(String.Format("Use QuickSeek: {0}", soundTouch.GetUseQuickSeek()));
            Global.Buzz.DCWriteLine(String.Format("Use AntiAliasing: {0}", soundTouch.GetUseAntiAliasing()));

            soundTouch.SetSampleRate(sampleRate);
            soundTouch.SetChannels(1);
            soundTouch.SetRateChange(changePercent);

            soundTouch.SetSequenceMs(sequenceMs);
            soundTouch.SetSeekWindowMs(seekWindowMs);
            soundTouch.SetOverlapMs(overlapMs);

            int totalInputSampleCount = 0;
            int totalOutputSampleCount = 0;
            float[] inputBuffer = new float[1000];
            int inputBufferFillLevel = 0;
            float[] outputBuffer = new float[1000];
            int outputBufferFillAmount = 0;

            do
            {
                bool endOfInput = totalInputSampleCount == sourceData.Length;

                if (inputBufferFillLevel == 0)
                {
                    // Refill input buffer
                    inputBufferFillLevel = Math.Min(1000, sourceData.Length - totalInputSampleCount);
                    Array.Copy(sourceData, totalInputSampleCount, inputBuffer, 0, inputBufferFillLevel);
                }

                soundTouch.PutSamples(inputBuffer, inputBufferFillLevel);
                totalInputSampleCount += inputBufferFillLevel;

                outputBufferFillAmount = Math.Min(1000, destData.Length - totalOutputSampleCount);
                outputBufferFillAmount = Math.Min(outputBufferFillAmount, soundTouch.NumberOfSamplesAvailable);
                outputBufferFillAmount = soundTouch.ReceiveSamples(outputBuffer, outputBufferFillAmount);

                Array.Copy(outputBuffer, 0, destData, totalOutputSampleCount, outputBufferFillAmount);

                totalOutputSampleCount += outputBufferFillAmount;

                inputBufferFillLevel = 0;
            }
            while (sourceData.Length - totalInputSampleCount > 0 || outputBufferFillAmount > 0);

            soundTouch.Dispose();

        }

        /// <summary>
        /// Returns sample frequency.
        /// </summary>
        /// <param name="wabetableWaveIndex"></param>
        /// <returns></returns>
        public int GetSampleFrequency(int wabetableWaveIndex)
        {
            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
            int sampleRate = 44100;

            lock (syncLock)
            {
                if (wt.Waves.Count > wabetableWaveIndex && wt.Waves[wabetableWaveIndex].Layers[0] != null)
                {
                    sampleRate = wt.Waves[wabetableWaveIndex].Layers[0].SampleRate;
                }
            }
            return sampleRate;
        }

        /// <summary>
        /// Is stereo sample?
        /// </summary>
        /// <param name="wabetableWaveIndex"></param>
        /// <returns></returns>
        public bool IsStereo(int wabetableWaveIndex)
        {
            IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;

            bool ret = true;

            lock (syncLock)
            {
                if (wabetableWaveIndex >= 0 && wt.Waves.Count > wabetableWaveIndex && wt.Waves[wabetableWaveIndex].Layers[0] != null)
                {
                    if (wt.Waves[wabetableWaveIndex].Layers[0].ChannelCount < 2)
                        ret = false;
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns playing position in pattern editor pattern when pattern is being played solo.
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public int GetPositionInPlayingPatternEditorPattern(out string patternName)
        {
            patternName = "";
            foreach (IPattern pat in host.Machine.Patterns)
            {
                if (pat.IsPlayingSolo && pat.PlayPosition >= 0)
                {
                    patternName = pat.Name;
                    double avgSamplesPerTick = (60.0 * host.MasterInfo.SamplesPerSec) / (host.MasterInfo.BeatsPerMin * host.MasterInfo.TicksPerBeat);
                    return (int)((pat.PlayPosition * avgSamplesPerTick + host.MasterInfo.PosInTick) / (double)PatternEvent.TimeBase);
                }
                else return -1;
            }
            return -1;
        }

        /// <summary>
        /// Returns playing position in pattern or -1.
        /// </summary>
        /// <param name="patternName"></param>
        /// <param name="seq"></param>
        /// <returns></returns>
        public int GetPositionInPattern(out string patternName, ISequence seq)
        {
            patternName = "";

            if (seq.Machine.Name == this.host.Machine.Name)
            {
                IPattern pat = seq.PlayingPattern;
                if (pat != null && pat.PlayPosition >= 0)
                {
                    patternName = pat.Name;
                    double avgSamplesPerTick = (60.0 * host.MasterInfo.SamplesPerSec) / (host.MasterInfo.BeatsPerMin * host.MasterInfo.TicksPerBeat);
                    return (int)(seq.PlayingPatternPosition * avgSamplesPerTick + host.MasterInfo.PosInTick);
                }
                else return -1;
            }
            return -1;
        }

        /// <summary>
        /// Returns plauing position in specific pattern or -1.
        /// </summary>
        /// <param name="patternName"></param>
        /// <returns></returns>
        public int GetPlayPositionInPatternName(string patternName)
        {
            foreach (IPattern pat in host.Machine.Patterns)
            {
                if (pat.Name == patternName && pat.PlayPosition >= 0)
                {
                    double avgSamplesPerTick = (60.0 * host.MasterInfo.SamplesPerSec) / (host.MasterInfo.BeatsPerMin * host.MasterInfo.TicksPerBeat);
                    return (int)((pat.PlayPosition * avgSamplesPerTick + host.MasterInfo.PosInTick) / (double)PatternEvent.TimeBase);
                }
            }
            return -1;
        }

        /// <summary>
        /// Return all sequences that are being played.
        /// </summary>
        /// <param name="machine"></param>
        /// <returns></returns>
        public ReadOnlyCollection<ISequence> GetPlayingSequences(IMachine machine)
        {
            List<ISequence> sequences = new List<ISequence>();
            foreach (ISequence seq in host.Machine.Graph.Buzz.Song.Sequences)
            {
                if (seq.Machine.Name == this.host.Machine.Name)
                {
                    IPattern pat = seq.PlayingPattern;
                    if (pat != null)
                    {
                        sequences.Add(seq);
                    }
                }
            }
            return sequences.AsReadOnly();
        }

        /// <summary>
        /// Linear volume ramp up for sample if sample is being cut.
        /// </summary>
        /// <param name="output"></param>
        /// <param name="start"></param>
        /// <param name="end"></param>
        public void SmoothCut(ref Sample[] output, int start, int end)
        {
            int pos = start;
            int direction = start < end ? 1 : -1;
            end = end < output.Length ? end : output.Length;

            float smoothMulStep = 1.0f / Math.Abs(end - start);
            float mul = 0.0f;

            while (pos != end)
            {
                output[pos].L *= mul;
                output[pos].R *= mul;
                mul += smoothMulStep;
                pos += direction;
            }
        }

        /// <summary>
        /// Get full file name for grag&drop file.
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="e"></param>
        /// <returns></returns>
        public bool GetFilename(out string filename, DragEventArgs e)
        {
            bool ret = true;
            filename = String.Empty;

            if ((e.AllowedEffects & DragDropEffects.Copy) == DragDropEffects.Copy)
            {
                Array data = ((IDataObject)e.Data).GetData("FileName") as Array;
                if (data != null)
                {
                    if ((data.Length == 1) && (data.GetValue(0) is String))
                    {
                        filename = ((string[])data)[0];
                        FileInfo fi = new FileInfo(filename);
                        filename = fi.FullName;
                    }
                }
            }
            return ret;
        }

        /// <summary>
        /// Create new AudioBlock pattern.
        /// </summary>
        /// <param name="patternLength"></param>
        /// <returns></returns>
        public string AddPattern(int patternLength)
        {
            string patternName = AudioBlock.AUDIO_BLOCK_PATTERN_BASE_NAME;
            for (int i = 0; i < 500; i++)
            {
                if (IsPatternNameAvailable(patternName + i.ToString("00")))
                {
                    patternName = patternName + i.ToString("00");
                    break;
                }
            }
            host.Machine.CreatePattern(patternName, patternLength);
            return patternName;
        }

        /// <summary>
        /// Get pattern object based on pattern name.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public IPattern GetPattern(string name)
        {
            IPattern ret = null;

            for (int i = 0; i < host.Machine.Patterns.Count; i++)
            {
                if (host.Machine.Patterns[i].Name == name)
                {
                    ret = host.Machine.Patterns[i];
                    break;
                }
            }
            return ret;
        }


        /// <summary>
        /// Calculate pattern lenght needed for sample.
        /// </summary>
        /// <param name="samples"></param>
        /// <param name="sampleRate"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public int CalculatePatternLength(int samples, int sampleRate, double offset)
        {
            double sampleCount = samples - offset * sampleRate;
            sampleCount = sampleCount < 0 ? 1 : sampleCount;
            double samplePlaySeconds = sampleCount / (double)sampleRate;
            double avgTicks = (samplePlaySeconds / 60.0) * (double)host.MasterInfo.BeatsPerMin * (double)host.MasterInfo.TicksPerBeat;
            double patternBlocks = avgTicks;

            int patternLength = (int)Math.Ceiling(patternBlocks);

            if (patternLength % PATTERN_LENGHT_STEP > 0)
            {
                patternLength = patternLength / PATTERN_LENGHT_STEP;
                patternLength = (patternLength + 1) * PATTERN_LENGHT_STEP;
            }

            return patternLength;
        }

        /// <summary>
        /// Return pattern lenght in seconds.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public double GetPatternLenghtInSeconds(string name)
        {
            double ret = 0;
            IPattern pat = GetPattern(name);
            if (pat != null)
            {
                double ticksPerSecond = (double)host.MasterInfo.BeatsPerMin / 60.0 * (double)host.MasterInfo.TicksPerBeat;
                ret = (double)pat.Length / ticksPerSecond;
            }

            return ret;
        }

        /// <summary>
        /// Note that variable name 'audioBlockIndex' is used in AudioBlock table and 'wavetableIndex' (or similar) in Buzz Wavetable.
        /// Don't mix them up!
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        public void SetPatternLength(int audioBlockIndex)
        {
            //lock (syncLock)
            {
                var wt = host.Machine.Graph.Buzz.Song.Wavetable;
                int waveIndex = MachineState.AudioBlockInfoTable[audioBlockIndex].WavetableIndex;

                // Don't set lenght if looping
                if (waveIndex > -1 && !MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled)
                {
                    IPattern pat = GetPattern(MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern);
                    if (pat != null)
                    {
                        pat.Length = CalculatePatternLength(wt.Waves[waveIndex].Layers[0].SampleCount, wt.Waves[waveIndex].Layers[0].SampleRate,
                            MachineState.AudioBlockInfoTable[audioBlockIndex].OffsetInMs / 1000.0 + MachineState.AudioBlockInfoTable[audioBlockIndex].OffsetInSeconds);
                        MachineState.AudioBlockInfoTable[audioBlockIndex].WavetableIndex = waveIndex;
                        RefreshBuzzViews();
                    }
                }
            }
        }

        /// <summary>
        /// Is pattern name available?
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public bool IsPatternNameAvailable(string name)
        {
            bool ret = true;

            for (int i = 0; i < host.Machine.Patterns.Count; i++)
            {
                if (host.Machine.Patterns[i].Name == name)
                {
                    ret = false;
                    break;
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns next available (empty) slot in Buzz Wavetable.
        /// </summary>
        /// <returns></returns>
        public int FindNextAvaialbeIndexInWavetable()
        {
            int ret = -1;
            var wt = host.Machine.Graph.Buzz.Song.Wavetable;
            for (int i = 0; i < wt.Waves.Count; i++)
            {
                if (wt.Waves[i] == null)
                {
                    ret = i;
                    break;
                }
            }
            return ret;
        }

        /// <summary>
        /// Returns longes sample in seconds. Offset is calcualted in the result.
        /// </summary>
        /// <returns></returns>
        public double GetLongestSampleInSecondsWithOffset()
        {
            double longest = 0;

            lock (syncLock)
            {
                IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
                for (int abIndex = 0; abIndex < this.MachineState.AudioBlockInfoTable.Length; abIndex++)
                {
                    int wavetableIndex = MachineState.AudioBlockInfoTable[abIndex].WavetableIndex;
                    if (wavetableIndex >= 0 && wt.Waves[wavetableIndex] != null)
                    {
                        double length = GetSampleLengthInSecondsWithOffset(abIndex);
                        if (length > longest)
                            longest = length;
                    }
                }
            }

            longest = longest == 0 ? 10 : longest;
            return longest;
        }

        /// <summary>
        /// Returns sample lenght in seconds.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <returns></returns>
        public double GetSampleLenghtInSeconds(int audioBlockIndex)
        {
            double length = 0;
            lock (syncLock)
            {
                if (MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled &&
                    MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
                {
                    length = GetPatternLenghtInSeconds(MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern);
                }
                else
                {
                    IWavetable wt = host.Machine.Graph.Buzz.Song.Wavetable;
                    int wavetableIndex = MachineState.AudioBlockInfoTable[audioBlockIndex].WavetableIndex;
                    if (wavetableIndex >= 0 && wt.Waves[wavetableIndex] != null)
                    {
                        length = wt.Waves[wavetableIndex].Layers[0].SampleCount / (double)wt.Waves[wavetableIndex].Layers[0].SampleRate;
                    }
                }
            }

            return length;
        }

        /// <summary>
        /// Returns sample lenght in seconds. Offset is calcualted in to the result.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <returns></returns>
        public double GetSampleLengthInSecondsWithOffset(int audioBlockIndex)
        {
            double length = GetSampleLenghtInSeconds(audioBlockIndex);
            if (MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled && MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
            {
                // Do nothing
            }
            else
            {
                // Add offset
                length -= (machineState.AudioBlockInfoTable[audioBlockIndex].OffsetInSeconds + machineState.AudioBlockInfoTable[audioBlockIndex].OffsetInMs / 1000);
                length = length < 0 ? 0 : length;
            }

            return length;
        }

        /// <summary>
        /// Small hack to notfity Buzz that user needs to save song when exiting Buzz.
        /// </summary>
        public void NotifyBuzzDataChanged()
        {
            int val = host.Machine.ParameterGroups[0].Parameters[0].GetValue(0);
            host.Machine.ParameterGroups[0].Parameters[0].SetValue(0, val);
        }

        // Helper interface
        public interface IEnvelopePoint
        {
            double TimeStamp { get; set; }
            double Value { get; set; }
        }


        // Here we save all the Volume Envelope time stamps and values. Could use struct, but class it is.
        public class VolumeEnvelopePoint : IEnvelopePoint
        {
            double timeStamp;
            double value;

            public double TimeStamp { get => timeStamp; set => timeStamp = value; }
            public double Value { get => value; set => this.value = value; }

            public VolumeEnvelopePoint()
            {
                TimeStamp = -1;
                Value = 0;
            }
        }

        public class PanEnvelopePoint : IEnvelopePoint
        {
            double timeStamp;
            double value;

            public double TimeStamp { get => timeStamp; set => timeStamp = value; }
            public double Value { get => value; set => this.value = value; }

            public PanEnvelopePoint()
            {
                TimeStamp = -1;
                Value = 0;
            }
        }

        // This is our data structure that is saved in to song.
        public class AudioBlockInfo
        {
            private int wavetableIndex;
            private double offsetInMs;
            private double offsetInSeconds;
            private string pattern;
            private float gain;
            private int color;

            private bool volumeEnvelopeEnabled;
            private int volumeEnvelopeMode;
            private List<VolumeEnvelopePoint> volumeEnvelopePoints;

            private bool panEnvelopeEnabled;
            private List<PanEnvelopePoint> panEnvelopePoints;

            private bool loopEnabled;
            private bool openLoop;
            private bool volumeEnvelopeCurves;
            private bool panEnvelopeCurves;

            public int WavetableIndex { get => wavetableIndex; set => wavetableIndex = value; }
            public string Pattern { get => pattern; set => pattern = value; }
            public double OffsetInMs { get => offsetInMs; set => offsetInMs = value; }
            public float Gain { get => gain; set => gain = value; }
            public double OffsetInSeconds { get => offsetInSeconds; set => offsetInSeconds = value; }
            public int Color { get => color; set => color = value; }
            public List<VolumeEnvelopePoint> VolumeEnvelopePoints { get => volumeEnvelopePoints; set => volumeEnvelopePoints = value; }
            public bool VolumeEnvelopeEnabled { get => volumeEnvelopeEnabled; set => volumeEnvelopeEnabled = value; }
            public int VolumeEnvelopeMode { get => volumeEnvelopeMode; set => volumeEnvelopeMode = value; }
            public bool LoopEnabled { get => loopEnabled; set => loopEnabled = value; }
            public bool OpenLoop { get => openLoop; set => openLoop = value; }
            public bool PanEnvelopeEnabled { get => panEnvelopeEnabled; set => panEnvelopeEnabled = value; }
            public List<PanEnvelopePoint> PanEnvelopePoints { get => panEnvelopePoints; set => panEnvelopePoints = value; }
            public bool VolumeEnvelopeCurves { get => volumeEnvelopeCurves; set => volumeEnvelopeCurves = value; }
            public bool PanEnvelopeCurves { get => panEnvelopeCurves; set => panEnvelopeCurves = value; }

            public AudioBlockInfo()
            {
                wavetableIndex = -1;
                offsetInMs = 0.0;
                offsetInSeconds = 0.0;
                pattern = "";
                gain = 1.0f;
                color = -1;

                loopEnabled = false;
                openLoop = false;

                volumeEnvelopeCurves = false;
                panEnvelopeCurves = false;

                VolumeEnvelopeEnabled = false;
                VolumeEnvelopeMode = 0;
                VolumeEnvelopePoints = new List<VolumeEnvelopePoint>();

                PanEnvelopeEnabled = false;
                PanEnvelopePoints = new List<PanEnvelopePoint>();
            }
        }

        public class State
        {
            public State()
            {
                for (int i = 0; i < NUMBER_OF_AUDIO_BLOCKS; i++)
                {
                    AudioBlockInfoTable[i] = new AudioBlockInfo();
                }

                waveViewOrientation = 0;
                overwriteSample = false;
            }	// NOTE: parameterless constructor is required by the xml serializer

            private AudioBlockInfo[] audioBlockInfoTable = new AudioBlockInfo[NUMBER_OF_AUDIO_BLOCKS];
            private int waveViewOrientation;
            private bool overwriteSample;

            public AudioBlockInfo[] AudioBlockInfoTable
            {
                get
                {
                    return audioBlockInfoTable;
                }
                set
                {
                    audioBlockInfoTable = value;

                    // Disable open loop until better solution is found.
                    foreach (AudioBlockInfo abi in audioBlockInfoTable)
                        abi.OpenLoop = false;
                }
            }

            public int WaveViewOrientation { get => waveViewOrientation; set => waveViewOrientation = value; }
            public bool OverwriteSample { get => overwriteSample; set => overwriteSample = value; }
        }

        State machineState = new State();
        public State MachineState           // a property called 'MachineState' gets automatically saved in songs and presets
        {
            get { return machineState; }
            set
            {
                machineState = value;
            }
        }

        /// <summary>
        /// Show Amiga influenced about box.
        /// </summary>
        private void ShowAboutWindow()
        {
            if (aboutWindow != null && aboutWindow.Visibility == Visibility.Visible)
                aboutWindow.Close();
                        
            aboutWindow = new AboutWindow(AUDIO_BLOCK_VERSION);
            aboutWindow.Show();
        }

        // Right click machine menu
        public IEnumerable<IMenuItem> Commands
        {
            get
            {
                yield return new MenuItemVM()
                {
                    Text = "Help",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => MessageBox.Show("AudioBlock turns patterns into audio blocks. Use Buzz Wavetable or drag & drop audio files to AudioBlock.")
                    }
                };

                yield return new MenuItemVM()
                {
                    Text = "About...",
                    Command = new SimpleCommand()
                    {
                        CanExecuteDelegate = p => true,
                        ExecuteDelegate = p => ShowAboutWindow()
                    }
                };
            }
        }

        public bool MachineInitialized { get; set; } = false;
        public AudioBlockGUI AudioBlockGUI { get; private set; }


        internal void SetGUI(AudioBlockGUI audioBlockGUI)
        {
            this.AudioBlockGUI = audioBlockGUI;
        }

        // Raise event when wave graphs need to be updated
        public event OnUpdateWaveGraph UpdateWaveGraph;
        public delegate void OnUpdateWaveGraph(AudioBlock ab, EventArgs e);

        /// <summary>
        /// Sends notification to registered callers.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        public void RaiseUpdateWaveGraphEvent(int audioBlockIndex)
        {
            EventArgsWaveUpdate e = new EventArgsWaveUpdate();
            e.AudioBlockIndex = audioBlockIndex;
            Application.Current.Dispatcher.BeginInvoke((Action)(() =>
            {
                UpdateWaveGraph(this, e);
            }));

        }

        /// <summary>
        /// Send notification toregisterd callers.
        /// </summary>
        /// <param name="audioBlockIndex"></param>
        /// <param name="type"></param>
        public void RaiseUpdateWaveGraphEvent(int audioBlockIndex, WaveUpdateEventType type)
        {
            EventArgsWaveUpdate e = new EventArgsWaveUpdate();
            e.AudioBlockIndex = audioBlockIndex;
            e.Type = type;
            if (UpdateWaveGraph != null)
                UpdateWaveGraph(this, e);
        }
    }

    // Event types to use for the UpdateWaveGraph event.
    public enum WaveUpdateEventType
    {
        Other,
        Drag,
        NewWaveAdded,
        PatternChanged,
        ColorChanged
    }

    // Our own event agrs class to be used with UpdateWaveGraph event.
    public class EventArgsWaveUpdate : EventArgs
    {
        private WaveUpdateEventType type = WaveUpdateEventType.Other;
        private int audioBlockIndex = -1;
        public int AudioBlockIndex { get => audioBlockIndex; set => audioBlockIndex = value; }
        internal WaveUpdateEventType Type { get => type; set => type = value; }
    }

    /// <summary>
    /// GUI Starts here.
    /// </summary>
    public class MachineGUIFactory : IMachineGUIFactory { public IMachineGUI CreateGUI(IMachineGUIHost host) { return new AudioBlockGUI(); } }
    public class AudioBlockGUI : UserControl, IMachineGUI
    {
        IMachine machine;
        AudioBlock audioBlockMachine;
        AudioBlockUIItem[] audioBlockUIItems;
        private Grid gridMain;
        private WaveView waveView;
        private ViewMode viewMode = ViewMode.Single;

        private MenuItem miOverwriteSample;

        internal WaveView WaveView { get => waveView; set => waveView = value; }


        public IMachine Machine
        {
            get { return machine; }
            set
            {
                machine = value;

                if (machine != null)
                {
                    audioBlockMachine = (AudioBlock)machine.ManagedMachine;
                    audioBlockMachine.SetGUI(this);

                    for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
                    {
                        audioBlockUIItems[i].SetAudioBlockMachine(audioBlockMachine, this);
                    }

                    // Properly initialize viewView content when it is first opened. Now just do the basics.
                    WaveView = new WaveView(audioBlockMachine, AudioBlock.WAVE_CANVAS_WIDTH, AudioBlock.WAVE_CANVAS_HEIGTH);

                    Grid.SetColumn(WaveView, 1);

                    WaveView.Init(); // Get the UpdateWaveGraph events after wavecanvases get it.
                    WaveView.Visibility = Visibility.Hidden;
                    gridMain.Children.Add(WaveView);

                    miOverwriteSample.IsChecked = audioBlockMachine.MachineState.OverwriteSample;

                    UpdateUI();
                }
            }
        }



        /// <summary>
        /// Fetch data from AudioBlock machine State and restore UI selections.
        /// </summary>
        public void UpdateUI()
        {
            for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
            {
                audioBlockUIItems[i].DisableEvents();
                audioBlockUIItems[i].PopulateData();
                audioBlockUIItems[i].RestoreSelection(
                    audioBlockMachine.MachineState.AudioBlockInfoTable[i].Pattern,
                    audioBlockMachine.MachineState.AudioBlockInfoTable[i].WavetableIndex,
                    audioBlockMachine.MachineState.AudioBlockInfoTable[i].OffsetInSeconds,
                    audioBlockMachine.MachineState.AudioBlockInfoTable[i].OffsetInMs,
                    audioBlockMachine.MachineState.AudioBlockInfoTable[i].Gain);
                audioBlockUIItems[i].EnableEvents();
            }
        }

        public AudioBlockGUI()
        {
            gridMain = new Grid() { VerticalAlignment = VerticalAlignment.Stretch, HorizontalAlignment = HorizontalAlignment.Stretch };
            ColumnDefinition gridCol1 = new ColumnDefinition();
            ColumnDefinition gridCol2 = new ColumnDefinition();
            gridMain.ColumnDefinitions.Add(gridCol1);
            gridMain.ColumnDefinitions.Add(gridCol2);

            var sp = new StackPanel() { };

            Grid.SetColumn(sp, 0);
            gridMain.Children.Add(sp);

            // Grid for controls. Maybe settings button in the future.
            Grid gridControls = new Grid { HorizontalAlignment = HorizontalAlignment.Stretch };
            gridControls.RowDefinitions.Add(new RowDefinition());
            gridControls.ColumnDefinitions.Add(new ColumnDefinition() { });

            Button bt = new Button() { Margin = new Thickness(8, 8, 20, 8), Content = new TextBlock() { Text = "Show / Hide Audio View", FontSize = 12 } };
            bt.Click += Bt_Click;
            Grid.SetColumn(bt, 0);

            ContextMenu menuOpt = new ContextMenu();
            miOverwriteSample = new MenuItem();
            miOverwriteSample.Header = "Overwrite Sample in Wavetable";
            miOverwriteSample.IsCheckable = true;
            miOverwriteSample.Checked += Mi_Checked;
            miOverwriteSample.Unchecked += Mi_Unchecked;
            menuOpt.Items.Add(miOverwriteSample);

            gridControls.Children.Add(bt);

            var spMainLabel = new StackPanel();
            spMainLabel.Orientation = Orientation.Horizontal;
            Label labelPattern = new Label() { Content = "Pattern", Width = 110 };
            spMainLabel.Children.Add(labelPattern);
            Label labelSampleNumber = new Label() { Content = "Wavetable Audio", Width = 306 };
            labelSampleNumber.ContextMenu = menuOpt;
            spMainLabel.Children.Add(labelSampleNumber);
            Label labelSampleOffsetSeconds = new Label() { Content = "Offset (sec)", Width = 88, ToolTip = "Positive value cuts, negative adds silence/delay." };
            spMainLabel.Children.Add(labelSampleOffsetSeconds);
            Label labelSampleOffsetMs = new Label() { Content = "Offset (ms)", Width = 88, ToolTip = "Positive value cuts, negative adds silence/delay." };
            spMainLabel.Children.Add(labelSampleOffsetMs);
            Label labelGain = new Label() { Content = "Vol", Width = 30, ToolTip = "Audio block gain." };
            spMainLabel.Children.Add(labelGain);

            sp.Children.Add(gridControls);
            sp.Children.Add(spMainLabel);

            gridCol1.Width = new GridLength(AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH);

            audioBlockUIItems = new AudioBlockUIItem[AudioBlock.NUMBER_OF_AUDIO_BLOCKS];
            for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
            {
                audioBlockUIItems[i] = new AudioBlockUIItem(i);
                sp.Children.Add(audioBlockUIItems[i]);
            }

            this.Content = gridMain;
            this.Loaded += AudioBlockGUI_Loaded;

            this.KeyDown += AudioBlockGUI_KeyDown;
        }



        private void Mi_Unchecked(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.MachineState.OverwriteSample = false;
        }

        private void Mi_Checked(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.MachineState.OverwriteSample = true;
        }

        private void AudioBlockGUI_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (WaveView.Visibility == Visibility.Visible && e.HeightChanged)
                WaveView.UpdateHeight(machine.ParameterWindow.ActualHeight - 200);
        }

        enum ViewMode
        {
            Single,
            Double
        }

        /// <summary>
        /// Change the AudioBlock UI between param view and param view + audio wave view.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Bt_Click(object sender, RoutedEventArgs e)
        {
            if (viewMode == ViewMode.Single)
            {
                viewMode = ViewMode.Double;

                machine.ParameterWindow.MinWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH + AudioBlock.AUDIO_BLOCK_GUI_WAVE_VIEW_MIN_WIDTH;
                machine.ParameterWindow.MaxWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH + 4 * AudioBlock.AUDIO_BLOCK_GUI_WAVE_VIEW_MIN_WIDTH;
                WaveView.UpdateScale();
                WaveView.Visibility = Visibility.Visible;

                // Init waves first time waveView is shown 
                if (!WaveView.IsShownBefore)
                {
                    for (int i = 0; i < AudioBlock.NUMBER_OF_AUDIO_BLOCKS; i++)
                    {
                        if (audioBlockMachine.MachineState.AudioBlockInfoTable[i].WavetableIndex != -1 && audioBlockMachine.MachineState.AudioBlockInfoTable[i].Pattern != "")
                        {
                            Brush brush = Utils.GetBrushForPattern(audioBlockMachine, i);

                            WaveView.AddWave(i, brush);
                        }
                    }

                    WaveView.ForceUpdateCanvases();
                    WaveView.IsShownBefore = true;
                }

                WaveView.UpdateGraph(); // Try to update heavy changes only when the UI is displayed. Aaaand this calls canvas GraphUpdate only if canvas flag is set.
            }
            else
            {
                viewMode = ViewMode.Single;

                WaveView.Visibility = Visibility.Hidden;
                machine.ParameterWindow.MinWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH;
                machine.ParameterWindow.MaxWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH;
            }
        }


        /// <summary>
        /// Do some resizing here.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AudioBlockGUI_Loaded(object sender, RoutedEventArgs e)
        {
            // Resize param window
            machine.ParameterWindow.MinWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH;
            machine.ParameterWindow.MaxWidth = AudioBlock.AUDIO_BLOCK_GUI_PARAM_WIDTH;
            machine.ParameterWindow.InvalidateVisual();

            machine.ParameterWindow.SizeChanged += AudioBlockGUI_SizeChanged;
        }


        private void AudioBlockGUI_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.Key == Key.Z && (Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control)
            {
                audioBlockMachine.WaveUndo.Undo();
            }
            else if (e.Key == Key.Y && (Keyboard.Modifiers & ModifierKeys.Control) == ModifierKeys.Control)
            {
                audioBlockMachine.WaveUndo.Redo();
            }
            e.Handled = true;
        }
    }
}
