﻿using Buzz.MachineInterface;
using BuzzGUI.Common;
using BuzzGUI.Interfaces;
using System;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Windows.Threading;

namespace WDE.AudioBlock
{
    /// <summary>
    /// WaveCanvas handles the actual wave drawing.
    /// </summary>
    public partial class WaveCanvas : Canvas
    {
        AudioBlock audioBlockMachine;
        private int audioBlockIndex;
        private Brush waveBrush;

        private static int MAXIMUN_SAMPLE_READ_SIZE = 1500;
        private static double DRAW_SCALE = 0.6;

        private double drawLengthInSeconds = 5.0;
        private double drawAccuracy = 1;
        private bool selectedWave;

        private bool updateGraphFlag = false;

        private ViewOrientationMode viewOrientationMode = ViewOrientationMode.Horizontal;

        private double slidingWindowOffsetSeconds = 0;

        private double playPositionLinePos = 0;
        private Line playPositionLine = null;

        private Canvas backgroundWaveCanvas;

        MenuItem menuItemQuality;
        DispatcherTimer dispatcherTimer;

        Envelopes envelopes;


        public Brush WaveBrush { get => waveBrush; set => waveBrush = value; }

        public double DrawLengthInSeconds
        {
            get => drawLengthInSeconds;

            set
            {
                drawLengthInSeconds = value;
                envelopes.DrawLengthInSeconds = drawLengthInSeconds;
            }
        }

        public bool SelectedWave
        {
            get { return selectedWave; }
            internal set
            {
                selectedWave = value;
                if (SelectedWave)
                {
                    this.Background = new SolidColorBrush(Utils.ChangeColorBrightness(Global.Buzz.ThemeColors["SE Pattern Box"], 0.4f));
                }
                else
                    this.Background = new SolidColorBrush(Global.Buzz.ThemeColors["SE Pattern Box"]);
            }
        }

        public double DrawAccuracy { get => drawAccuracy; set => drawAccuracy = value; }
        public int AudioBlockIndex { get => audioBlockIndex; set => audioBlockIndex = value; }
        public bool UpdateGraphFlag { get => updateGraphFlag; set => updateGraphFlag = value; }
        internal ViewOrientationMode ViewOrientationMode { get => viewOrientationMode; set => viewOrientationMode = value; }
        public double SlidingWindowOffsetSeconds { get => slidingWindowOffsetSeconds; set => slidingWindowOffsetSeconds = value; }
        public MenuItem MenuItemQuality { get => menuItemQuality; set => menuItemQuality = value; }

        private MenuItem miLooping;
        private MenuItem miLoopEnabled;
        // private MenuItem miOpenLoopEnabled;
        // private MenuItem miOpenLoop;

        public WaveCanvas(AudioBlock audioBlockMachine, int audioBlockIndex)
        {
            InitializeComponent();

            SelectedWave = false;

            WaveBrush = Brushes.Red;

            this.audioBlockMachine = audioBlockMachine;
            this.AudioBlockIndex = audioBlockIndex;

            backgroundWaveCanvas = new Canvas() { SnapsToDevicePixels = true };
            this.SnapsToDevicePixels = true;

            this.Background = Utils.CreateLGBackgroundBrush(ViewOrientationMode == ViewOrientationMode.Vertical);

            this.Loaded += WaveCanvas_Loaded;
            audioBlockMachine.UpdateWaveGraph += AudioBlockMachine_UpdateWaveGraph;

            dispatcherTimer = new DispatcherTimer();
            dispatcherTimer.Interval = new TimeSpan(0, 0, 0, 0, 1000 / 60);
            dispatcherTimer.Tick += DispatcherTimer_Tick;
            dispatcherTimer.Start();

            playPositionLine = Utils.CreateLine(this, -1, -1, -1, -1, new SolidColorBrush(Global.Buzz.ThemeColors["SE Song Position"]), 1);

            ContextMenu cmWaveCanvas = new ContextMenu() { Margin = new Thickness(4, 4, 4, 4) };

            miLooping = new MenuItem();
            miLooping.Header = "Looping";
            cmWaveCanvas.Items.Add(miLooping);
            miLoopEnabled = new MenuItem();
            miLoopEnabled.Header = "Enabled";
            miLoopEnabled.IsCheckable = true;
            miLoopEnabled.IsChecked = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled;
            miLoopEnabled.Checked += Mi_Checked_Looping;
            miLoopEnabled.Unchecked += Mi_Unchecked_Looping;
            miLooping.Items.Add(miLoopEnabled);

            // Open loop disabled until I figure out other approach            
            /*
            miOpenLoop = new MenuItem();
            miOpenLoop.Header = "Loop Repeat";
            miLooping.Items.Add(miOpenLoop);

            miOpenLoopEnabled = new MenuItem();
            miOpenLoopEnabled.Header = "Open Loop";
            miOpenLoopEnabled.IsCheckable = true;
            miOpenLoopEnabled.IsEnabled = true;
            miOpenLoopEnabled.IsChecked = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop;
            miOpenLoopEnabled.Checked += Mi_Checked_OpenLoop;
            miOpenLoopEnabled.Unchecked += Mi_Unchecked_OpenLoop;
            miOpenLoop.Items.Add(miOpenLoopEnabled);

            MenuItem mi = new MenuItem();
            mi.Header = "Loop Length...";            
            miOpenLoop.Items.Add(mi);
            */

            this.ContextMenu = cmWaveCanvas;
            UpdateLoopMenu();

            this.Unloaded += WaveCanvas_Unloaded;

            menuItemQuality = new MenuItem();
            menuItemQuality.Header = "Low Quality";
            menuItemQuality.IsCheckable = true;
            cmWaveCanvas.Items.Add(menuItemQuality);
            EnableEvents();

            envelopes = new Envelopes(audioBlockMachine, this) { SnapsToDevicePixels = true };

            // This will create a bitmap copy of backgroundWaveCanvas and that is used as cache to speed up rendering.
            backgroundWaveCanvas.CacheMode = new BitmapCache() { EnableClearType = false, SnapsToDevicePixels = false };

            this.MouseWheel += WaveCanvas_MouseWheel;
        }

        private void WaveCanvas_MouseWheel(object sender, System.Windows.Input.MouseWheelEventArgs e)
        {
            if (Keyboard.IsKeyDown(Key.LeftCtrl) || Keyboard.IsKeyDown(Key.RightCtrl))
            {
                Slider sliderZoom = audioBlockMachine.AudioBlockGUI.WaveView.SliderZoom;
                sliderZoom.Value += sliderZoom.Maximum / e.Delta;
                e.Handled = true;
            }
            if (Keyboard.IsKeyDown(Key.LeftShift) || Keyboard.IsKeyDown(Key.RightShift))
            {
                Slider sliderWindow = audioBlockMachine.AudioBlockGUI.WaveView.SliderWindow;
                sliderWindow.Value += sliderWindow.Maximum / e.Delta;
                e.Handled = true;
            }
        }

        private void UpdateLoopMenu()
        {
            if (miLoopEnabled.IsChecked)
            {
                for (int i = 1; i < miLooping.Items.Count; i++)
                    ((MenuItem)miLooping.Items[i]).IsEnabled = true;
            }
            else
            {
                for (int i = 1; i < miLooping.Items.Count; i++)
                    ((MenuItem)miLooping.Items[i]).IsEnabled = false;
            }
        }

        private void Mi_Unchecked_OpenLoop(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.SetOpenLoopState(audioBlockIndex, false);
            audioBlockMachine.AudioBlockGUI.WaveView.UpdateScale();
            audioBlockMachine.UpdateEnvPoints(audioBlockIndex);
            audioBlockMachine.NotifyBuzzDataChanged();
            this.UpdateGraph();
        }

        private void Mi_Checked_OpenLoop(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.SetOpenLoopState(audioBlockIndex, true);
            audioBlockMachine.AudioBlockGUI.WaveView.UpdateScale();
            audioBlockMachine.UpdateEnvPoints(audioBlockIndex);
            audioBlockMachine.NotifyBuzzDataChanged();
            this.UpdateGraph();
        }

        private void Mi_Unchecked_Looping(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.SetLoopingState(audioBlockIndex, false);
            audioBlockMachine.AudioBlockGUI.WaveView.UpdateScale();
            audioBlockMachine.NotifyBuzzDataChanged();
            audioBlockMachine.UpdateEnvPoints(audioBlockIndex);
            audioBlockMachine.SetPatternLength(this.audioBlockIndex);
            UpdateLoopMenu();
            this.UpdateGraph();
        }

        private void Mi_Checked_Looping(object sender, RoutedEventArgs e)
        {
            audioBlockMachine.SetLoopingState(audioBlockIndex, true);
            audioBlockMachine.AudioBlockGUI.WaveView.UpdateScale();
            audioBlockMachine.NotifyBuzzDataChanged();
            audioBlockMachine.UpdateEnvPoints(audioBlockIndex);
            UpdateLoopMenu();
            this.UpdateGraph();
        }

        private void WaveCanvas_Unloaded(object sender, RoutedEventArgs e)
        {
            DisableEvents();
            dispatcherTimer.Stop();
        }

        /// <summary>
        /// Update play position line.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DispatcherTimer_Tick(object sender, EventArgs e)
        {
            if (this.Visibility == Visibility.Visible && (Global.Buzz.Playing || Global.Buzz.Recording))
            {
                playPositionLinePos = this.audioBlockMachine.GetPlayPositionInPatternName(audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern);
                if (playPositionLinePos >= 0)
                {
                    playPositionLinePos = UpdatePlayPosIfLooped(playPositionLinePos);

                    UpdatePlayLine();
                    playPositionLine.Visibility = Visibility.Visible;
                }
                else
                {
                    playPositionLinePos = UpdatePlayPosIfLooped(playPositionLinePos);

                    UpdatePlayLine();
                    playPositionLine.Visibility = Visibility.Hidden;
                }
            }
        }

        /// <summary>
        /// Update play position if looping AudioBlock item.
        /// </summary>
        /// <param name="playPos"></param>
        /// <returns></returns>
        private double UpdatePlayPosIfLooped(double playPos)
        {
            double ret = playPos;

            if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled &&
                !audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
            {
                lock (AudioBlock.syncLock)
                {

                    IWavetable wt = audioBlockMachine.host.Machine.Graph.Buzz.Song.Wavetable;

                    int sampleNum = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].WavetableIndex;
                    double step = ((double)audioBlockMachine.GetSampleFrequency(sampleNum)) / ((double)Global.Buzz.SelectedAudioDriverSampleRate);

                    string patternName = audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Pattern;
                    double offset = (audioBlockMachine.GetOffsetForPatternInMs(patternName) / 1000.0) * (double)wt.Waves[sampleNum].Layers[0].SampleRate;


                    double loopStart = wt.Waves[sampleNum].Layers[0].LoopStart;
                    double loopEnd = wt.Waves[sampleNum].Layers[0].LoopEnd;
                    double loopLength = loopEnd - loopStart;

                    if (playPos * step > loopStart - offset)
                    {
                        ret = (((playPos * step) - (loopStart - offset)) % loopLength);
                        ret += loopStart - offset;
                        ret /= step;
                    }
                }
            }

            return ret;
        }

        public void DisableEvents()
        {
            MenuItemQuality.Checked -= Mi_Canvas_Quality_Checked;
            MenuItemQuality.Unchecked -= Mi_Canvas_Quality_Unchecked;
        }

        public void EnableEvents()
        {
            MenuItemQuality.Unchecked += Mi_Canvas_Quality_Unchecked;
            MenuItemQuality.Checked += Mi_Canvas_Quality_Checked;
        }

        private void Mi_Canvas_Quality_Unchecked(object sender, RoutedEventArgs e)
        {
            StackPanel spWaves = this.audioBlockMachine.AudioBlockGUI.WaveView.SpWaves;
            for (int i = 0; i < spWaves.Children.Count; i++)
            {
                WaveCanvas wc = (WaveCanvas)spWaves.Children[i];
                wc.DisableEvents();
                wc.MenuItemQuality.IsChecked = false;
                wc.EnableEvents();
                wc.DrawAccuracy = 1;
                wc.UpdateGraph();
            }
        }

        /// <summary>
        /// Changing drawing quality does not affect the audio and in the end might not speed up the drawing that much...
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Mi_Canvas_Quality_Checked(object sender, RoutedEventArgs e)
        {
            StackPanel spWaves = this.audioBlockMachine.AudioBlockGUI.WaveView.SpWaves;
            for (int i = 0; i < spWaves.Children.Count; i++)
            {
                WaveCanvas wc = (WaveCanvas)spWaves.Children[i];
                wc.DisableEvents();
                wc.MenuItemQuality.IsChecked = true;
                wc.EnableEvents();
                wc.DrawAccuracy = 2;
                wc.UpdateGraph();
            }
        }

        private double CalculatePlayPosInCanvas()
        {
            double ret;
            if (ViewOrientationMode == ViewOrientationMode.Horizontal)
            {
                double sampleLength = ((double)Global.Buzz.SelectedAudioDriverSampleRate) * DrawLengthInSeconds;
                ret = (playPositionLinePos - SlidingWindowOffsetSeconds * (double)Global.Buzz.SelectedAudioDriverSampleRate) * (this.Height / (double)sampleLength);
            }
            else
            {
                double sampleLength = ((double)Global.Buzz.SelectedAudioDriverSampleRate) * DrawLengthInSeconds;
                ret = (playPositionLinePos - SlidingWindowOffsetSeconds * (double)Global.Buzz.SelectedAudioDriverSampleRate) * (this.Width / (double)sampleLength);
            }

            return ret;
        }

        public void UpdatePlayLine()
        {
            if (ViewOrientationMode == ViewOrientationMode.Horizontal)
            {
                double posInCanvas = CalculatePlayPosInCanvas();
                Utils.UpdateLine(playPositionLine, 0, posInCanvas, this.Width, posInCanvas);
            }
            else
            {
                double posInCanvas = CalculatePlayPosInCanvas();
                Utils.UpdateLine(playPositionLine, posInCanvas, 0, posInCanvas, this.Height);
            }
        }

        /// <summary>
        /// Do not update graph here because it is expensive. Rather raise a flag and do it OnRender.
        /// </summary>
        /// <param name="ab"></param>
        /// <param name="e"></param>
        private void AudioBlockMachine_UpdateWaveGraph(AudioBlock ab, EventArgs e)
        {
            EventArgsWaveUpdate eau = (EventArgsWaveUpdate)e;
            if (this.AudioBlockIndex == eau.AudioBlockIndex)
            {
                if (eau.Type == WaveUpdateEventType.ColorChanged)
                {
                    int color = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Color;
                    if (color > -1)
                        WaveBrush = new SolidColorBrush(Color.FromRgb((byte)((color & 0xFF0000) >> 16), (byte)((color & 0xFF00) >> 8), (byte)(color & 0xFF)));
                }

                if (this.IsVisible)
                {
                    UpdateGraph();
                }
                else
                {
                    UpdateGraphFlag = true;
                }
            }
        }

        /// <summary>
        /// Don't do heavy redrawing until allowed.
        /// </summary>
        /// <param name="dc"></param>
        protected override void OnRender(DrawingContext dc)
        {
            if (UpdateGraphFlag)
            {
                UpdateGraph();
                UpdateGraphFlag = false;
            }

            base.OnRender(dc);
        }

        /// <summary>
        /// Handles the actual drawing of the whole graph view...
        /// </summary>
        public void UpdateGraph()
        {
            Line playLineTmp = playPositionLine;

            this.Children.Clear();

            backgroundWaveCanvas.Children.Clear();
            backgroundWaveCanvas.Width = this.Width;
            backgroundWaveCanvas.Height = this.Height;

            if (ViewOrientationMode == ViewOrientationMode.Horizontal)
                DrawVerticalWave(backgroundWaveCanvas);
            else
                DrawHorizontalWave(backgroundWaveCanvas);

            this.Children.Add(backgroundWaveCanvas);

            Utils.DrawText(this, audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Pattern);

            envelopes.ViewOrientationMode = this.ViewOrientationMode;
            this.Children.Add(envelopes);
            envelopes.Draw();

            // Play position line is updated on top of wave
            if (playLineTmp != null)
                this.Children.Add(playLineTmp);

            Thread.Sleep(10); // Drawing is very heavy operation so give time to other threads.
            UpdatePlayLine();
        }

        private void WaveCanvas_Loaded(object sender, RoutedEventArgs e)
        {
            UpdateGraph();
        }

        internal void ChangeOrientation(ViewOrientationMode orientationMode)
        {
            this.ViewOrientationMode = orientationMode;
            double tmpSize = this.Width;
            this.Width = this.Height;
            this.Height = tmpSize;

            envelopes.ViewOrientationMode = this.ViewOrientationMode;
            envelopes.Width = this.Width;
            envelopes.Height = this.Height;
        }


        /// <summary>
        /// All this is a bit hacky. I wanted to reuse some functions in AudioBlock.cs and that forces some weirdish code. At some point add proper fucntions
        /// that do exactly what is needed...
        /// 
        /// Also Canvas could be separated in to different class to embed multiple waves to window.
        /// </summary>
        /// <param name="canvas"></param>
        public void DrawVerticalWave(Canvas canvas)
        {
            IWavetable wt = audioBlockMachine.host.Machine.Graph.Buzz.Song.Wavetable;

            int sampleNum = audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].WavetableIndex;
            if (sampleNum < 0 || wt.Waves[sampleNum] == null)
                return;

            string patternName = audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Pattern;
            bool isStereo = audioBlockMachine.IsStereo(sampleNum);
            double offsetInSamples = (audioBlockMachine.GetOffsetForPatternInMs(patternName) / 1000.0 + SlidingWindowOffsetSeconds) * (double)wt.Waves[sampleNum].Layers[0].SampleRate;
            double sampleLength;

            sampleLength = ((double)Global.Buzz.SelectedAudioDriverSampleRate) * DrawLengthInSeconds;

            //if (SelectedWave)
            //{
            //    this.Background = new SolidColorBrush(Utils.ChangeColorBrightness(Global.Buzz.ThemeColors["SE Pattern Box"], 0.4f));
            //}

            this.Background = Utils.CreateLGBackgroundBrush(ViewOrientationMode == ViewOrientationMode.Vertical);

            double drawScale = DRAW_SCALE * audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Gain;

            // Only draw every x pixel to speed up graph
            double readStep = (double)sampleLength / (canvas.Height / DrawAccuracy);

            int readSize = (int)readStep < MAXIMUN_SAMPLE_READ_SIZE ? (int)readStep : MAXIMUN_SAMPLE_READ_SIZE;
            readSize = readSize == 0 ? 1 : readSize;

            double readPos = 0;

            Brush strokeBrush = new SolidColorBrush(Color.FromArgb(0x70, 0, 0, 0));

            Polygon LeftWave = new Polygon();
            LeftWave.Height = canvas.Height;
            LeftWave.Fill = WaveBrush;
            LeftWave.Stroke = strokeBrush;
            LeftWave.StrokeThickness = 1;

            Polygon RightWave = new Polygon();
            RightWave.Fill = WaveBrush;
            RightWave.Stroke = strokeBrush;
            RightWave.StrokeThickness = 1;

            int numPoints = (int)(canvas.Height / DrawAccuracy);
            Point[] pointsL = new Point[numPoints];
            Point[] pointsR = new Point[numPoints];

            Sample[] output = new Sample[readSize];

            for (int i = 0; i < numPoints; i++)
            {
                for (int j = 0; j < readSize; j++)
                {
                    output[j] = new Sample(0.0f, 0.0f);
                }

                GetSamples(AudioBlockIndex, ref output, readSize, readPos);

                readPos += readStep;

                pointsL[i] = GetMinMaxL(ref output, readSize);
                if (isStereo)
                    pointsR[i] = GetMinMaxR(ref output, readSize);
            }

            double leftWidth = canvas.Width;
            double leftCenter = canvas.Width / 2.0;
            double rightWidth = canvas.Width / 2.0;
            double rightCenter = canvas.Width * 3.0 / 4.0;

            if (isStereo)
            {
                leftWidth /= 2.0;
                leftCenter /= 2.0;
            }

            for (int y = 0; y < numPoints; y++)
            {
                Point point = new Point();
                point.X = (pointsL[y].Y * leftWidth * drawScale) + leftCenter;
                point.Y = y * DrawAccuracy;
                LeftWave.Points.Add(point);
            }

            for (int y = numPoints - 1; y >= 0; y--)
            {
                Point point = new Point();
                point.X = (pointsL[y].X * leftWidth * drawScale) + leftCenter;
                point.Y = y * DrawAccuracy;
                LeftWave.Points.Add(point);
            }

            if (isStereo)
            {
                for (int y = 0; y < numPoints; y++)
                {
                    Point point = new Point();
                    point.X = (pointsR[y].Y * rightWidth * drawScale) + rightCenter;
                    point.Y = y * DrawAccuracy;
                    RightWave.Points.Add(point);
                }

                for (int y = numPoints - 1; y >= 0; y--)
                {
                    Point point = new Point();
                    point.X = (pointsR[y].X * rightWidth * drawScale) + rightCenter;
                    point.Y = y * DrawAccuracy;
                    RightWave.Points.Add(point);
                }
            }

            Brush tickBrush = new SolidColorBrush(Utils.ChangeColorBrightness(Global.Buzz.ThemeColors["SE Pattern Box"], -0.3f));
            DrawTickLinesForVerWave(canvas, offsetInSamples, sampleNum, sampleLength, tickBrush);

            canvas.Children.Add(LeftWave);
            if (isStereo)
                canvas.Children.Add(RightWave);

            Utils.DrawBox(backgroundWaveCanvas);

            if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled)
                DrawLoopPointsForVerticalWave(canvas, sampleNum, sampleLength, Brushes.Yellow);

            DrawOffsetForVerWave(canvas, offsetInSamples, sampleNum, sampleLength, strokeBrush);
        }

        public void DrawHorizontalWave(Canvas canvas)
        {
            IWavetable wt = audioBlockMachine.host.Machine.Graph.Buzz.Song.Wavetable;

            int sampleNum = audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].WavetableIndex;
            if (sampleNum < 0)
                return;

            string patternName = audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Pattern;
            bool isStereo = audioBlockMachine.IsStereo(sampleNum);
            double offset = (audioBlockMachine.GetOffsetForPatternInMs(patternName) / 1000.0 + SlidingWindowOffsetSeconds) * (double)wt.Waves[sampleNum].Layers[0].SampleRate;
            double sampleLength;

            sampleLength = ((double)Global.Buzz.SelectedAudioDriverSampleRate) * DrawLengthInSeconds;

            //if (SelectedWave)
            //{
            //    this.Background = new SolidColorBrush(Utils.ChangeColorBrightness(Global.Buzz.ThemeColors["SE Pattern Box"], 0.4f));
            //}

            this.Background = Utils.CreateLGBackgroundBrush(ViewOrientationMode == ViewOrientationMode.Vertical);

            double drawScale = DRAW_SCALE * audioBlockMachine.MachineState.AudioBlockInfoTable[AudioBlockIndex].Gain;

            // Only draw every x pixel to speed up graph
            double readStep = (double)sampleLength / (canvas.Width / DrawAccuracy);

            int readSize = (int)readStep < MAXIMUN_SAMPLE_READ_SIZE ? (int)readStep : MAXIMUN_SAMPLE_READ_SIZE;
            readSize = readSize == 0 ? 1 : readSize;

            double readPos = 0;

            Polygon LeftWave = new Polygon();
            LeftWave.Fill = WaveBrush;

            Brush strokeBrush = new SolidColorBrush(Color.FromArgb(0x70, 0, 0, 0));
            LeftWave.Stroke = strokeBrush;
            LeftWave.StrokeThickness = 1;

            Polygon RightWave = new Polygon();
            RightWave.Fill = WaveBrush;

            RightWave.Stroke = strokeBrush;
            RightWave.StrokeThickness = 1;

            int numPoints = (int)(canvas.Width / DrawAccuracy);
            Point[] pointsL = new Point[numPoints];
            Point[] pointsR = new Point[numPoints];

            Sample[] output = new Sample[readSize];

            for (int i = 0; i < numPoints; i++)
            {
                for (int j = 0; j < readSize; j++)
                {
                    output[j] = new Sample(0.0f, 0.0f);
                }

                GetSamples(AudioBlockIndex, ref output, readSize, readPos);

                readPos += readStep;

                pointsL[i] = GetMinMaxL(ref output, readSize);
                if (isStereo)
                    pointsR[i] = GetMinMaxR(ref output, readSize);
            }

            // ToDo: Get Overall Min/Max to scale the wave

            double leftHeight = canvas.Height;
            double leftCenter = canvas.Height / 2.0;
            double rightHeight = canvas.Height / 2.0;
            double rightCenter = canvas.Height * 3.0 / 4.0;

            if (isStereo)
            {
                leftHeight /= 2.0;
                leftCenter /= 2.0;
            }

            for (int x = 0; x < numPoints; x++)
            {
                Point point = new Point();
                point.Y = (pointsL[x].Y * leftHeight * drawScale) + leftCenter;
                point.X = x * DrawAccuracy;
                LeftWave.Points.Add(point);
            }

            for (int x = numPoints - 1; x >= 0; x--)
            {
                Point point = new Point();
                point.Y = (pointsL[x].X * leftHeight * drawScale) + leftCenter;
                point.X = x * DrawAccuracy;
                LeftWave.Points.Add(point);
            }

            if (isStereo)
            {
                for (int x = 0; x < numPoints; x++)
                {
                    Point point = new Point();
                    point.Y = (pointsR[x].Y * rightHeight * drawScale) + rightCenter;
                    point.X = x * DrawAccuracy;
                    RightWave.Points.Add(point);
                }

                for (int x = numPoints - 1; x >= 0; x--)
                {
                    Point point = new Point();
                    point.Y = (pointsR[x].X * rightHeight * drawScale) + rightCenter;
                    point.X = x * DrawAccuracy;
                    RightWave.Points.Add(point);
                }
            }

            Brush tickBrush = new SolidColorBrush(Utils.ChangeColorBrightness(Global.Buzz.ThemeColors["SE Pattern Box"], -0.3f));
            DrawTickLinesForHorWave(canvas, offset, sampleNum, sampleLength, tickBrush);

            canvas.Children.Add(LeftWave);
            if (isStereo)
                canvas.Children.Add(RightWave);

            Utils.DrawBox(backgroundWaveCanvas);

            if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled)
                DrawLoopPointsForHorizontalWave(canvas, sampleNum, sampleLength, Brushes.Yellow);

            DrawOffsetForHorWave(canvas, offset, sampleNum, sampleLength, strokeBrush);
        }

        private void DrawOffsetForVerWave(Canvas canvas, double offset, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            // Draw Offset Line
            if (offset < 0)
            {
                double step = ((double)audioBlockMachine.GetSampleFrequency(sampleNum)) / ((double)Global.Buzz.SelectedAudioDriverSampleRate);

                double offsetY = canvas.Height * (-offset / (sampleLength));
                offsetY /= step;

                Line myLine = new Line();

                myLine.Stroke = strokeBrush;
                myLine.SnapsToDevicePixels = true;
                myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
                myLine.StrokeThickness = 1;

                myLine.StrokeDashArray = new DoubleCollection() { 4, 2 };

                myLine.X1 = 0;
                myLine.Y1 = offsetY;
                myLine.X2 = canvas.Width - 1;
                myLine.Y2 = offsetY;

                canvas.Children.Add(myLine);
            }
        }

        private void DrawLoopPointsForVerticalWave(Canvas canvas, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            IWavetable wt = audioBlockMachine.host.Machine.Graph.Buzz.Song.Wavetable;
            double step = ((double)audioBlockMachine.GetSampleFrequency(sampleNum)) / ((double)Global.Buzz.SelectedAudioDriverSampleRate);

            string patternName = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern;
            double offset = (audioBlockMachine.GetOffsetForPatternInMs(patternName) / 1000.0 + SlidingWindowOffsetSeconds) * (double)wt.Waves[sampleNum].Layers[0].SampleRate;

            double loopStart = wt.Waves[sampleNum].Layers[0].LoopStart;
            double loopEnd = wt.Waves[sampleNum].Layers[0].LoopEnd;

            double startY = canvas.Height * ((loopStart - offset) / (sampleLength));
            startY /= step;

            double endY = 0;


            if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled &&
                audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
            {
                endY = audioBlockMachine.GetPatternLenghtInSeconds(audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern) *
                    (double)audioBlockMachine.GetSampleFrequency(sampleNum);
                endY = canvas.Height * endY / sampleLength;
            }
            else
            {
                endY = canvas.Height * ((loopEnd - offset) / (sampleLength));
            }

            endY /= step;

            Line myLine = new Line();

            myLine.Stroke = strokeBrush;
            myLine.SnapsToDevicePixels = true;
            myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
            myLine.StrokeThickness = 1.3;

            myLine.StrokeDashArray = new DoubleCollection() { 3, 2 };

            myLine.X1 = 0;
            myLine.Y1 = startY;
            myLine.X2 = canvas.Width - 1;
            myLine.Y2 = startY;

            canvas.Children.Add(myLine);

            // End
            myLine = new Line();

            myLine.Stroke = strokeBrush;
            myLine.SnapsToDevicePixels = true;
            myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
            myLine.StrokeThickness = 1.3;

            myLine.StrokeDashArray = new DoubleCollection() { 3, 2 };

            myLine.X1 = 0;
            myLine.Y1 = endY;
            myLine.X2 = canvas.Width - 1;
            myLine.Y2 = endY;

            canvas.Children.Add(myLine);
        }

        private void DrawLoopPointsForHorizontalWave(Canvas canvas, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            IWavetable wt = audioBlockMachine.host.Machine.Graph.Buzz.Song.Wavetable;
            double step = ((double)audioBlockMachine.GetSampleFrequency(sampleNum)) / ((double)Global.Buzz.SelectedAudioDriverSampleRate);

            string patternName = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern;
            double offset = (audioBlockMachine.GetOffsetForPatternInMs(patternName) / 1000.0 + SlidingWindowOffsetSeconds) * (double)wt.Waves[sampleNum].Layers[0].SampleRate;

            double loopStart = wt.Waves[sampleNum].Layers[0].LoopStart;
            double loopEnd = wt.Waves[sampleNum].Layers[0].LoopEnd;

            double startX = canvas.Width * ((loopStart - offset) / (sampleLength));
            startX /= step;

            double endX = 0;

            if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled &&
                            audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
            {
                endX = audioBlockMachine.GetPatternLenghtInSeconds(audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern) *
                    (double)audioBlockMachine.GetSampleFrequency(sampleNum);
                endX = canvas.Width * endX / sampleLength;
            }
            else
            {
                endX = canvas.Width * ((loopEnd - offset) / (sampleLength));
            }

            endX /= step;

            Line myLine = new Line();

            myLine.Stroke = strokeBrush;
            myLine.SnapsToDevicePixels = true;
            myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
            myLine.StrokeThickness = 1.3;

            myLine.StrokeDashArray = new DoubleCollection() { 3, 2 };

            myLine.X1 = startX;
            myLine.Y1 = 0;
            myLine.X2 = startX;
            myLine.Y2 = canvas.Height - 1;

            canvas.Children.Add(myLine);

            // End
            myLine = new Line();

            myLine.Stroke = strokeBrush;
            myLine.SnapsToDevicePixels = true;
            myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
            myLine.StrokeThickness = 1.3;

            myLine.StrokeDashArray = new DoubleCollection() { 3, 2 };

            myLine.X1 = endX;
            myLine.Y1 = 0;
            myLine.X2 = endX;
            myLine.Y2 = canvas.Height - 1;

            canvas.Children.Add(myLine);
        }

        private void DrawTickLinesForVerWave(Canvas canvas, double offset, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            double numberOfTicks = sampleLength / (double)audioBlockMachine.host.MasterInfo.SamplesPerTick;
            double pixelsPerTick = canvas.Height / numberOfTicks;

            double offsetInPixels = (SlidingWindowOffsetSeconds * (double)audioBlockMachine.host.MasterInfo.TicksPerSec) * pixelsPerTick;

            double start = pixelsPerTick - offsetInPixels;

            if (start < 0)
                start = pixelsPerTick - (-start % pixelsPerTick);

            int ticksPerBeat = audioBlockMachine.host.MasterInfo.TicksPerBeat;
            int ticking = ((int)(offsetInPixels / pixelsPerTick)) % ticksPerBeat + 1;

            if (pixelsPerTick > 10)
            {
                for (double y = start; y < canvas.Height; y += pixelsPerTick)
                {
                    Line myLine = new Line();

                    myLine.Stroke = strokeBrush;
                    myLine.SnapsToDevicePixels = true;
                    myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
                    myLine.StrokeThickness = 1;

                    if (ticking % ticksPerBeat == 0)
                        myLine.StrokeThickness = 2;

                    ticking++;

                    myLine.X1 = 0;
                    myLine.Y1 = y;
                    myLine.X2 = canvas.Width - 1;
                    myLine.Y2 = y;

                    canvas.Children.Add(myLine);
                }
            }
        }

        private void DrawTickLinesForHorWave(Canvas canvas, double offset, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            double numberOfTicks = sampleLength / (double)audioBlockMachine.host.MasterInfo.SamplesPerTick;
            double pixelsPerTick = canvas.Width / numberOfTicks;

            double offsetInPixels = (SlidingWindowOffsetSeconds * (double)audioBlockMachine.host.MasterInfo.TicksPerSec) * pixelsPerTick;

            double start = pixelsPerTick - offsetInPixels;

            if (start < 0)
                start = pixelsPerTick - (-start % pixelsPerTick);

            int ticksPerBeat = audioBlockMachine.host.MasterInfo.TicksPerBeat;
            int ticking = ((int)(offsetInPixels / pixelsPerTick)) % ticksPerBeat + 1;

            if (pixelsPerTick > 10)
            {
                for (double x = start; x < canvas.Width; x += pixelsPerTick)
                {
                    Line myLine = new Line();

                    myLine.Stroke = strokeBrush;
                    myLine.SnapsToDevicePixels = true;
                    myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
                    myLine.StrokeThickness = 1;

                    if (ticking % ticksPerBeat == 0)
                        myLine.StrokeThickness = 2;

                    ticking++;

                    myLine.X1 = x;
                    myLine.Y1 = 0;
                    myLine.X2 = x;
                    myLine.Y2 = canvas.Height - 1;

                    canvas.Children.Add(myLine);
                }
            }
        }

        private void DrawOffsetForHorWave(Canvas canvas, double offset, int sampleNum, double sampleLength, Brush strokeBrush)
        {
            offset += SlidingWindowOffsetSeconds;
            // Draw Offset Line
            if (offset < 0)
            {
                double step = ((double)audioBlockMachine.GetSampleFrequency(sampleNum)) / ((double)Global.Buzz.SelectedAudioDriverSampleRate);

                double offsetX = canvas.Width * (-offset / (sampleLength));
                offsetX /= step;

                Line myLine = new Line();

                myLine.Stroke = strokeBrush;
                myLine.SnapsToDevicePixels = true;
                myLine.SetValue(RenderOptions.EdgeModeProperty, EdgeMode.Aliased);
                myLine.StrokeThickness = 1;

                myLine.StrokeDashArray = new DoubleCollection() { 4, 2 };

                myLine.X1 = offsetX;
                myLine.Y1 = 0;
                myLine.X2 = offsetX;
                myLine.Y2 = canvas.Height - 1;

                canvas.Children.Add(myLine);
            }
        }

        private Point GetMinMaxL(ref Sample[] output, int readSize)
        {
            float min = 100000;
            float max = -100000;

            for (int i = 0; i < readSize; i++)
            {
                min = Math.Min(output[i].L, min);
                max = Math.Max(output[i].L, max);
            }
            return new Point(min, max);
        }

        private Point GetMinMaxR(ref Sample[] output, int readSize)
        {
            float min = 100000;
            float max = -100000;

            for (int i = 0; i < readSize; i++)
            {
                min = Math.Min(output[i].R, min);
                max = Math.Max(output[i].R, max);
            }
            return new Point(min, max);
        }

        private void GetSamples(int audioBlockWave, ref Sample[] output, int readSize, double pos)
        {

            double playPositionInSample = pos;

            int sampleIndex = audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockWave].WavetableIndex;
            if (sampleIndex >= 0)
            {
                double step = 1.0;

                step = audioBlockMachine.GetSampleFrequency(sampleIndex) / (double)Global.Buzz.SelectedAudioDriverSampleRate;

                Sample[] outputTmp = new Sample[(int)(Math.Ceiling(readSize * step))];
                double offset = (double)audioBlockMachine.GetSampleFrequency(sampleIndex) * (SlidingWindowOffsetSeconds + audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockWave].OffsetInSeconds +
                    audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockWave].OffsetInMs / 1000.0);

                playPositionInSample *= step;
                playPositionInSample += offset;

                double patternLenght = 0.0;
                bool cutIfLooped = false;
                // Don't draw beyond loop
                if (audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].LoopEnabled &&
                    audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].OpenLoop)
                {
                    playPositionInSample = audioBlockMachine.UpdatePlayPosIfLooped(playPositionInSample, sampleIndex, audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern);
                    patternLenght = audioBlockMachine.GetPatternLenghtInSeconds(audioBlockMachine.MachineState.AudioBlockInfoTable[audioBlockIndex].Pattern) *
                        (double)Global.Buzz.SelectedAudioDriverSampleRate;
                    cutIfLooped = true;
                }

                int outputOffset = 0;

                // Offest in the beginning...
                if (playPositionInSample < 0)
                {
                    for (int i = 0; i < readSize; i++)
                    {
                        playPositionInSample += step;
                        outputOffset++;
                        if (playPositionInSample >= 0)
                            break;
                    }
                }

                if (cutIfLooped && (pos > patternLenght))
                {
                    // Pattern ended --> fill zeroes
                    for (int i = 0; i < readSize - outputOffset; i++)
                    {
                        output[i + outputOffset].L = 0;
                        output[i + outputOffset].R = 0;
                    }
                }
                else
                {
                    audioBlockMachine.GetSamples(sampleIndex, ref outputTmp, outputTmp.Length, (int)(playPositionInSample), false);

                    double stepper = 0.0;
                    for (int i = 0; i < readSize - outputOffset; i++)
                    {
                        output[i + outputOffset].L = outputTmp[(int)stepper].L;
                        output[i + outputOffset].R = outputTmp[(int)stepper].R;
                        stepper += step;
                    }
                }
            }
        }
    }
}
